"""
Input/Output operations on resources.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 27, 2012.
"""
from collections import OrderedDict
import os
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile

from pyramid.compat import NativeIO
from pyramid.compat import iteritems_
from pyramid.compat import text_type
from pyramid.compat import urlparse

from everest.compat import open_text
from everest.entities.utils import get_entity_class
from everest.mime import CsvMime
from everest.mime import MimeTypeRegistry
from everest.repositories.memory.cache import EntityCacheMap
from everest.representers.utils import as_representer
from everest.resources.attributes import get_resource_class_attribute_names
from everest.resources.attributes import get_resource_class_attributes
from everest.resources.attributes import is_resource_class_collection_attribute
from everest.resources.attributes import is_resource_class_terminal_attribute
from everest.resources.staging import StagingAggregate
from everest.resources.staging import create_staging_collection
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from everest.resources.utils import provides_member_resource
from pygraph.algorithms.sorting import topological_sorting # pylint: disable=E0611,F0401
from pygraph.classes.digraph import digraph # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['ConnectedResourcesSerializer',
           'ResourceGraph',
           'build_resource_dependency_graph',
           'build_resource_graph',
           'dump_resource',
           'dump_resource_to_files',
           'dump_resource_to_zipfile',
           'find_connected_resources',
           'get_collection_filename',
           'get_collection_name',
           'get_read_collection_path',
           'get_write_collection_path',
           'load_collection_from_file',
           'load_collection_from_stream',
           'load_collection_from_url',
           'load_into_collection_from_file',
           'load_into_collection_from_stream',
           'load_into_collection_from_url',
           'load_into_collections_from_zipfile',
           ]


def load_into_collection_from_stream(collection, stream, content_type):
    """
    Loads resources from the given resource data stream (of the specified MIME
    content type) into the given collection resource.
    """
    rpr = as_representer(collection, content_type)
    with stream:
        data_el = rpr.data_from_stream(stream)
    rpr.resource_from_data(data_el, resource=collection)


def load_collection_from_stream(resource, stream, content_type):
    """
    Creates a new collection for the registered resource and calls
    `load_into_collection_from_stream` with it.
    """
    coll = create_staging_collection(resource)
    load_into_collection_from_stream(coll, stream, content_type)
    return coll


def load_into_collection_from_file(collection, filename,
                                   content_type=None):
    """
    Loads resources from the specified file into the given collection
    resource.

    If no content type is provided, an attempt is made to look up the
    extension of the given filename in the MIME content type registry.
    """
    if content_type is None:
        ext = os.path.splitext(filename)[1]
        try:
            content_type = MimeTypeRegistry.get_type_for_extension(ext)
        except KeyError:
            raise ValueError('Could not infer MIME type for file extension '
                             '"%s".' % ext)
    load_into_collection_from_stream(collection, open(filename, 'rU'),
                                     content_type)


def load_collection_from_file(resource, filename, content_type=None):
    """
    Creates a new collection for the registered resource and calls
    `load_into_collection_from_file` with it.
    """
    coll = create_staging_collection(resource)
    load_into_collection_from_file(coll, filename,
                                   content_type=content_type)
    return coll


def load_into_collection_from_url(collection, url, content_type=None):
    """
    Loads resources from the representation contained in the given URL into
    the given collection resource.

    :returns: collection resource
    """
    parsed = urlparse.urlparse(url)
    scheme = parsed.scheme # pylint: disable=E1101
    if scheme == 'file':
        # Assume a local path.
        load_into_collection_from_file(collection,
                                       parsed.path, # pylint: disable=E1101
                                       content_type=content_type)
    else:
        raise ValueError('Unsupported URL scheme "%s".' % scheme)


def load_collection_from_url(resource, url, content_type=None):
    """
    Creates a new collection for the registered resource and calls
    `load_into_collection_from_url` with it.
    """
    coll = create_staging_collection(resource)
    load_into_collection_from_url(coll, url, content_type=content_type)
    return coll


class DecodingStream(object):
    """
    Helper class that iterates over a bytes stream yielding strings.
    """
    def __init__(self, stream, encoding='utf-8'):
        self.__stream = stream
        self.__encoding = encoding

    def __enter__(self):
        return self.__stream.__enter__()

    def __exit__(self, *args, **kw):
        self.__stream.__exit__(*args, **kw)

    def __iter__(self):
        for line in self.__stream:
            yield text_type(line, self.__encoding)


def load_into_collections_from_zipfile(collections, zipfile):
    """
    Loads resources contained in the given ZIP archive into each of the
    given collections.

    The ZIP file is expected to contain a list of file names obtained with
    the :func:`get_collection_filename` function, each pointing to a file
    of zipped collection resource data.

    :param collections: sequence of collection resources
    :param str zipfile: ZIP file name
    """
    with ZipFile(zipfile) as zipf:
        names = zipf.namelist()
        name_map = dict([(os.path.splitext(name)[0], index)
                         for (index, name) in enumerate(names)])
        for coll in collections:
            coll_name = get_collection_name(coll)
            index = name_map.get(coll_name)
            if index is None:
                continue
            coll_fn = names[index]
            ext = os.path.splitext(coll_fn)[1]
            try:
                content_type = \
                    MimeTypeRegistry.get_type_for_extension(ext)
            except KeyError:
                raise ValueError('Could not infer MIME type for file '
                                 'extension "%s".' % ext)
            # Strings are always written as UTF-8 encoded byte strings when
            # the zip file is created, so we have to wrap the iterator into
            # a decoding step.
            coll_data = DecodingStream(zipf.open(coll_fn, 'r'))
            load_into_collection_from_stream(coll,
                                             coll_data,
                                             content_type)


def dump_resource(resource, stream, content_type=None):
    """
    Dumps the given resource to the given stream using the specified MIME
    content type (defaults to CSV).
    """
    if content_type is None:
        content_type = CsvMime
    rpr = as_representer(resource, content_type)
    rpr.to_stream(resource, stream)


def build_resource_dependency_graph(resource_classes,
                                    include_backrefs=False):
    """
    Builds a graph of dependencies among the given resource classes.

    The dependency graph is a directed graph with member resource classes as
    nodes. An edge between two nodes represents a member or collection
    attribute.

    :param resource_classes: resource classes to determine interdependencies
      of.
    :type resource_classes: sequence of registered resources.
    :param bool include_backrefs: flag indicating if dependencies
      introduced by back-references (e.g., a child resource referencing its
      parent) should be included in the dependency graph.
    """
    def visit(mb_cls, grph, path, incl_backrefs):
        for attr_name in get_resource_class_attribute_names(mb_cls):
            if is_resource_class_terminal_attribute(mb_cls, attr_name):
                continue
            child_descr = getattr(mb_cls, attr_name)
            child_mb_cls = get_member_class(child_descr.attr_type)
            # We do not follow cyclic references back to a resource class
            # that is last in the path.
            if len(path) > 0 and child_mb_cls is path[-1] \
               and not incl_backrefs:
                continue
            if not grph.has_node(child_mb_cls):
                grph.add_node(child_mb_cls)
                path.append(mb_cls)
                visit(child_mb_cls, grph, path, incl_backrefs)
                path.pop()
            if not grph.has_edge((mb_cls, child_mb_cls)):
                grph.add_edge((mb_cls, child_mb_cls))
    dep_grph = digraph()
    for resource_class in resource_classes:
        mb_cls = get_member_class(resource_class)
        if not dep_grph.has_node(mb_cls):
            dep_grph.add_node(mb_cls)
            visit(mb_cls, dep_grph, [], include_backrefs)
    return dep_grph


def build_resource_graph(resource, dependency_graph=None):
    """
    Traverses the graph of resources that is reachable from the given
    resource.

    If a resource dependency graph is given, links to other resources are
    only followed if the dependency graph has an edge connecting the two
    corresponding resource classes; otherwise, a default graph is built
    which ignores all direct cyclic resource references.

    :resource: a :class:`everest.resources.MemberResource` instance.
    :returns: a :class:`ResourceGraph` instance representing the graph of
        resources reachable from the given resource.
    """
    def visit(rc, grph, dep_grph):
        mb_cls = type(rc)
        attr_map = get_resource_class_attributes(mb_cls)
        for attr_name, attr in iteritems_(attr_map):
            if is_resource_class_terminal_attribute(mb_cls, attr_name):
                continue
            # Only follow the resource attribute if the dependency graph
            # has an edge here.
            child_mb_cls = get_member_class(attr.attr_type)
            if not dep_grph.has_edge((mb_cls, child_mb_cls)):
                continue
            child_rc = getattr(rc, attr_name)
            if is_resource_class_collection_attribute(mb_cls, attr_name):
                for child_mb in child_rc:
                    if not grph.has_node(child_mb): # Ignore cyclic references.
                        grph.add_node(child_mb)
                        grph.add_edge((rc, child_mb))
                        visit(child_mb, grph, dep_grph)
            else: # Member.
                if not grph.has_node(child_rc): # Ignore cyclic references.
                    grph.add_node(child_rc)
                    grph.add_edge((rc, child_rc))
                    visit(child_rc, grph, dep_grph)
    if  dependency_graph is None:
        dependency_graph = build_resource_dependency_graph(
                                            [get_member_class(resource)])
    graph = ResourceGraph()
    if provides_member_resource(resource):
        rcs = [resource]
    else:
        rcs = resource
    for rc in rcs:
        graph.add_node(rc)
        visit(rc, graph, dependency_graph)
    return graph


def find_connected_resources(resource, dependency_graph=None):
    """
    Collects all resources connected to the given resource and returns a
    dictionary mapping member resource classes to new collections containing
    the members found.
    """
    # Build a resource_graph.
    resource_graph = \
                build_resource_graph(resource,
                                     dependency_graph=dependency_graph)
    entity_map = OrderedDict()
    for mb in topological_sorting(resource_graph):
        mb_cls = get_member_class(mb)
        ents = entity_map.get(mb_cls)
        if ents is None:
            ents = []
            entity_map[mb_cls] = ents
        ents.append(mb.get_entity())
    return entity_map


class ResourceGraph(digraph):
    """
    Specialized digraph for resource instances.

    Nodes are resources, edges represent relationships between resources.
    Since resources are wrapper objects generated on the fly, the presence
    of a resource in the graph is determined by its underlying entity, using
    the entity class and its ID as a key.
    """
    def __init__(self):
        digraph.__init__(self)
        self.__entities = set()

    def add_node(self, node, attrs=None):
        digraph.add_node(self, node, attrs=attrs)
        key = self.__make_key(node)
        self.__entities.add(key)

#    def del_node(self, node):
#        digraph.del_node(self, node)
#        key = self.__make_key(node)
#        self.__entities.remove(key)

    def has_node(self, node):
        return self.__make_key(node) in self.__entities

    def __make_key(self, obj):
        ent = obj.get_entity()
        return (type(ent), ent.id)


class ConnectedResourcesSerializer(object):
    """
    Serializer for a graph of connected resources.
    """
    def __init__(self, content_type, dependency_graph=None):
        """
        :param content_type: MIME content type to use for representations
        :type content_type: object implementing
            :class:`everest.interfaces.IMime`.
        :param dependency_graph: graph determining which resource connections
            to follow when the graph of connected resources for a given
            resource is built.
        """
        self.__content_type = content_type
        self.__dependency_graph = dependency_graph

    def __collect(self, resource):
        ent_cls = get_entity_class(resource)
        coll_cls = get_collection_class(resource)
        cache = EntityCacheMap()
        agg = StagingAggregate(ent_cls, cache)
        coll = coll_cls.create_from_aggregate(agg)
        coll.add(resource)
        return dict([(get_member_class(ent_cls),
                      coll.get_root_collection(ent_cls))
                     for ent_cls in cache.keys()])

    def to_strings(self, resource):
        """
        Dumps the all resources reachable from the given resource to a map of
        string representations using the specified content_type (defaults
        to CSV).

        :returns: dictionary mapping resource member classes to string
            representations
        """
        collections = self.__collect(resource)
        # Build a map of representations.
        rpr_map = OrderedDict()
        for (mb_cls, coll) in iteritems_(collections):
            strm = NativeIO('w')
            dump_resource(coll, strm, content_type=self.__content_type)
            rpr_map[mb_cls] = strm.getvalue()
        return rpr_map

    def to_files(self, resource, directory):
        """
        Dumps the given resource and all resources linked to it into a set of
        representation files in the given directory.
        """
        collections = self.__collect(resource)
        for (mb_cls, coll) in iteritems_(collections):
            fn = get_write_collection_path(mb_cls,
                                           self.__content_type,
                                           directory=directory)
            with open_text(os.path.join(directory, fn)) as strm:
                dump_resource(coll, strm, content_type=self.__content_type)

    def to_zipfile(self, resource, zipfile):
        """
        Dumps the given resource and all resources linked to it into the given
        ZIP file.
        """
        rpr_map = self.to_strings(resource)
        with ZipFile(zipfile, 'w') as zipf:
            for (mb_cls, rpr_string) in iteritems_(rpr_map):
                fn = get_collection_filename(mb_cls, self.__content_type)
                zipf.writestr(fn, rpr_string, compress_type=ZIP_DEFLATED)


def dump_resource_to_files(resource, content_type=None, directory=None):
    """
    Convenience function. See
    :meth:`everest.resources.io.ConnectedResourcesSerializer.to_files` for
    details.

    If no directory is given, the current working directory is used.
    The given context type defaults to CSV.
    """
    if directory is None:
        directory = os.getcwd() # pragma: no cover
    if content_type is None:
        content_type = CsvMime
    srl = ConnectedResourcesSerializer(content_type)
    srl.to_files(resource, directory=directory)


def dump_resource_to_zipfile(resource, zipfile, content_type=None):
    """
    Convenience function. See
    :meth:`everest.resources.io.ConnectedResourcesSerializer.to_zipfile` for
    details.

    The given context type defaults to CSV.
    """
    if content_type is None:
        content_type = CsvMime
    srl = ConnectedResourcesSerializer(content_type)
    srl.to_zipfile(resource, zipfile)


def get_collection_name(rc_class):
    coll_cls = get_member_class(rc_class)
    collection_name = coll_cls.relation.split('/')[-1]
    return "%s-collection" % collection_name


def get_collection_filename(rc_class, content_type=None):
    if content_type is None:
        content_type = CsvMime
    return "%s%s" % (get_collection_name(rc_class),
                     content_type.file_extension)


def get_write_collection_path(collection_class, content_type, directory=None):
    if directory is None:
        directory = os.getcwd() # pragma: no cover
    coll_fn = get_collection_filename(collection_class, content_type)
    return os.path.join(directory, coll_fn)


def get_read_collection_path(collection_class, content_type, directory=None):
    if directory is None:
        directory = os.getcwd() # pragma: no cover
    coll_fn = get_collection_filename(collection_class, content_type)
    fn = os.path.join(directory, coll_fn)
    if os.path.isfile(fn):
        result = fn
    else:
        result = None
    return result



"""

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 21, 2012.
"""
from everest.resources.storing import build_resource_dependency_graph
from everest.resources.utils import get_member_class
from everest.testing import TestCaseWithConfiguration
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityGrandchild
from everest.tests.complete_app.interfaces import IMyEntityParent

__docformat__ = 'reStructuredText en'
__all__ = ['ResourceGraphTestCase',
           ]


class ResourceGraphTestCase(TestCaseWithConfiguration):
    package_name = 'everest.tests.complete_app'

    def set_up(self):
        TestCaseWithConfiguration.set_up(self)
        self.config.load_zcml('configure_no_rdb.zcml')
        self._interfaces = [IMyEntityParent, IMyEntity, IMyEntityChild,
                            IMyEntityGrandchild]

    def test_dependency_graph(self):
        grph = build_resource_dependency_graph(self._interfaces)
        self.assert_equal(len(grph.nodes()), 4)
        entity_mb_cls = get_member_class(IMyEntity)
        entity_parent_mb_cls = get_member_class(IMyEntityParent)
        entity_child_mb_cls = get_member_class(IMyEntityChild)
        entity_grandchild_mb_cls = get_member_class(IMyEntityGrandchild)
        # Entity Parent resource deps should be empty.
        self.assert_equal(grph.neighbors(entity_parent_mb_cls), [])
        # Entity Child has Grandchild.
        self.assert_equal(grph.neighbors(entity_child_mb_cls),
                          [entity_grandchild_mb_cls])
        # Entity Grandchild has Child, but we ignore backreferences.
        self.assert_equal(grph.neighbors(entity_grandchild_mb_cls),
                          [])
        # Entity has Parent and Child.
        self.assert_equal(set(grph.neighbors(entity_mb_cls)),
                          set([entity_parent_mb_cls, entity_child_mb_cls]))



"""
Custom resource object tree traversal.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 4, 2011.
"""
from collections import MutableSequence
from collections import MutableSet
from everest.attributes import get_attribute_cardinality
from everest.attributes import is_terminal_attribute
from everest.constants import CARDINALITY_CONSTANTS
from everest.constants import RELATION_OPERATIONS
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.constants import RESOURCE_KINDS
from everest.interfaces import IDataTraversalProxyAdapter
from everest.interfaces import IDataTraversalProxyFactory
from everest.resources.interfaces import IResource
from everest.traversalpath import TraversalPath
from logging import getLogger as get_logger
from pyramid.compat import itervalues_
from pyramid.threadlocal import get_current_registry
from pyramid.traversal import ResourceTreeTraverser
from zope.interface import implementer # pylint: disable=E0611,F0401
#from everest.resources.staging import create_staging_collection

__docformat__ = 'reStructuredText en'
__all__ = ['ConvertingDataTraversalProxyMixin',
           'DataSequenceTraversalProxy',
           'DataTraversalProxy',
           'DataTraversalProxyAdapter',
           'DataTraversalProxyFactory',
           'SourceTargetDataTreeTraverser',
           'SuffixResourceTraverser',
           ]


class SuffixResourceTraverser(ResourceTreeTraverser):
    """
    A custom resource tree traverser that allows us to specify the
    representation for resources with a suffix as in
    `http://everest/myobjects.csv`.

    Rather than to reproduce the functionality of the `__call__` method, we
    check if base part of the current view name (`myobjects` in the example)
    can be retrieved as a child resource from the context. If yes, we set the
    context to the resource and the view name to the extension part of the
    current view name (`csv` in the example); if no, nothing is changed.
    """
    def __call__(self, request):
        system = ResourceTreeTraverser.__call__(self, request)
        context = system['context']
        view_name = system['view_name']
        if IResource.providedBy(context) and '.' in view_name: # pylint: disable=E1101
            rc_name, repr_name = view_name.split('.')
            try:
                child_rc = context[rc_name]
            except KeyError:
                pass
            else:
                if IResource.providedBy(child_rc): # pylint: disable=E1101
                    system['context'] = child_rc
                    system['view_name'] = repr_name
        return system


class DataSequenceTraversalProxy(object):
    """
    Simple wrapper for a sequence of data traversal proxies.
    """
    #: Constant indicating that this proxy is for collection resource data.
    proxy_for = RESOURCE_KINDS.COLLECTION

    def __init__(self, proxies):
        self.__proxies = proxies

    def __iter__(self):
        return iter(self.__proxies)


class DataTraversalProxy(object):
    """
    Abstract base class for data tree traversal proxies.

    By providing a uniform interface to the nodes of data trees
    encountered during tree traversal, this proxy makes it possible to use
    different data structures as source or target for the traversal.
    """
    #: Constant indicating that this proxy is for member resource data.
    proxy_for = RESOURCE_KINDS.MEMBER

    def __init__(self, data, accessor, relationship_direction):
        """
        :param data: root of the data tree to traverse.
        :param relationship_direction: constant indicating which relation
          direction(s) to consider for managing references.
        """
        super(DataTraversalProxy, self).__init__()
        self.relationship_direction = relationship_direction
        self._data = data
        self._accessor = accessor
        self._relationships = {}

    def get_relationship_attributes(self):
        """
        Returns an iterator over the relationship attributes (i.e.,
        non-terminal attributes) of the proxied data.

        :returns: iterator yielding objects implementing
          :class:`everest.resources.interfaces.IResourceAttribute`.
        """
        for attr in self._attribute_iterator():
            if not is_terminal_attribute(attr):
                yield attr

    def get_matching(self, source_id):
        """
        Returns a matching target object for the given source ID.
        """
        value = self._accessor.get_by_id(source_id)
        if not value is None:
            reg = get_current_registry()
            prx_fac = reg.getUtility(IDataTraversalProxyFactory)
            prx = prx_fac.make_proxy(value,
                                     self._accessor,
                                     self.relationship_direction)
        else:
            prx = None
        return prx

    @property
    def update_attribute_value_items(self):
        """
        Returns an iterator of items for an attribute value map to use for
        an UPDATE operation.

        The iterator ignores collection attributes as these are processed
        implicitly by the traversal algorithm.

        :returns: iterator yielding tuples with objects implementing
          :class:`everest.resources.interfaces.IResourceAttribute` as the
          first and the proxied attribute value as the second argument.
        """
        for attr in self._attribute_iterator():
            if attr.kind != RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
                try:
                    attr_val = self._get_proxied_attribute_value(attr)
                except AttributeError:
                    continue
                else:
                    yield (attr, attr_val)

    def set_relationship(self, attribute, relationship):
        """
        Sets the given domain relationship object for the given resource
        attribute.
        """
        self._relationships[attribute.entity_attr] = relationship

    def get_relationship(self, attribute):
        """
        Returns the domain relationship object for the given resource
        attribute.
        """
        return self._relationships[attribute.entity_attr]

    def get_attribute_proxy(self, attribute):
        """
        Returns a traversal proxy (cardinality ONE) or an iterable sequence
        data traversal proxy (cardinality MANY) for the specified relation
        attribute value of the proxied data.

        :raises ValueError: If :param:`attribute` is a terminal attribute.
        """
        attr_val = self._get_relation_attribute_value(attribute)
        if attr_val is None:
            prx = None
        else:
            if not self._accessor is None:
                acc = self._make_accessor(attribute.attr_type)
            else:
                acc = None
            reg = get_current_registry()
            prx_fac = reg.getUtility(IDataTraversalProxyFactory)
            prx = prx_fac.make_proxy(attr_val, acc,
                                     self.relationship_direction,
                                     options=
                                        self._get_proxy_options(attribute))
        return prx

    def do_traverse(self):
        """
        Checks if this proxy should be traversed.
        """
        return True

    def _get_proxy_options(self, attribute): # pylint:disable=W0613
        """
        Returns custom constructor options to pass to new proxies constructed
        from this one through :method:`get_attribute_proxy`. This default
        implementation returns an emtpy dictionary.
        """
        return {}

    def _get_relatee(self, attribute):
        """
        Returns the relatee for the given relation attribute.

        :raises AttributeError: If no relatee for the given attribute has
          been set.
        """
        try:
            rel = self._relationships[attribute.entity_attr]
        except KeyError:
            raise AttributeError(attribute)
        else:
            return rel.relatee

    def get_id(self):
        """
        Returns the ID of the proxied data.

        :returns: Numeric or string ID.
        """
        raise NotImplementedError('Abstract method.')

    def _get_entity_type(self):
        """
        Returns the entity type of the proxied data.

        :returns: Type object (subclass of
          :class:`everest.entities.base.Entity`)
        """
        raise NotImplementedError('Abstract method.')

    def get_entity(self):
        """
        Returns the proxied data as a domain object.

        :returns: Instance of :class:`everest.entities.base.Entity`.
        """
        raise NotImplementedError('Abstract method.')

    def _attribute_iterator(self):
        """
        Returns an iterator over all proxied resource attributes.
        """
        raise NotImplementedError('Abstract method.')

    def _get_relation_attribute_value(self, attribute):
        """
        Returns the value for the given relation attribute from the proxied
        data. Depending on the implementation, a sequence of data items may
        be returned for attributes of cardinality MANY.

        :raises AttributeError: If the underlying data does not have the
            given relation attribute set.
        """
        raise NotImplementedError('Abstract method.')

    def _get_proxied_attribute_value(self, attribute):
        """
        Returns the value for the given attribute from the proxied data.

        :raises AttributeError: If the proxied data do not have the specified
          attribute.
        """
        raise NotImplementedError('Abstract method.')

    def _make_accessor(self, value_type):
        """
        Creates a new target accessor for the given value.
        """
        raise NotImplementedError('Abstract method.')

    def __str__(self):
        return "%s(id=%s)" % (self._data.__class__.__name__, self.get_id())

    def __hash__(self):
        """
        The hash value is built either from the ID and the entity type of
        the proxied data or, if the ID is None, from the runtime object
        ID.
        """
        data_id = self.get_id()
        return data_id is None and id(self._data) \
               or hash((self._get_entity_type(), data_id))

    def __eq__(self, other):
        """
        Equality is determined by comparing hash values.
        """
        return self.__hash__() == hash(other)


@implementer(IDataTraversalProxyFactory)
class DataTraversalProxyFactory(object):
    """
    Factory for data traversal proxies.
    """
    def make_source_proxy(self, data, options=None):
        """
        Returns a data traversal proxy for the given source data.

        This is a convenience factory function to use when manually setting
        up a data traversal proxy for source data.

        :raises ValueError: If there is no adapter for the given traversal
          data and the data is not iterable.
        """
        return self.__make_proxy('make_source_proxy', data, (),
                                 dict(options=options))

    def make_target_proxy(self, data, accessor, manage_back_references=True,
                          options=None):
        """
        Returns a data traversal proxy for the given target data.

        This is a convenience factory function to use when manually setting
        up a data traversal proxy for target data.

        :raises ValueError: If there is no adapter for the given traversal
          data and the data is not iterable.
        """
        return self.__make_proxy('make_target_proxy', data, (accessor,),
                                 dict(manage_back_references=
                                            manage_back_references,
                                      options=options))

    def make_proxy(self, data, accessor, relationship_direction,
                   options=None):
        """
        Returns a data traversal proxy for the given data.

        This is the generic factory function that can be used to create
        both source and target data traversal proxies.

        :raises ValueError: If there is no adapter for the given traversal
          data and the data is not iterable.
        """
        return self.__make_proxy('make_proxy', data,
                                 (accessor, relationship_direction),
                                 dict(options=options))

    def __make_proxy(self, method_name, data, args, options):
        # We first check if we have a registered adapter for the given data;
        # if not, we assume it is a mutable sequence or set; if that fails,
        # we raise a ValueError.
        reg = get_current_registry()
        adp = reg.queryAdapter(data, IDataTraversalProxyAdapter)
        if not adp is None:
            prx = getattr(adp, method_name)(*args, **options)
        else:
            if not isinstance(data, (MutableSequence, MutableSet)):
                # Assuming an iterable is not enough here.
                raise ValueError('Invalid data type for traversal: %s.'
                                 % type(data))
            else:
                prxs = []
                for item in data:
                    adp = reg.queryAdapter(item, IDataTraversalProxyAdapter)
                    if adp is None:
                        raise ValueError('Invalid data type for traversal: '
                                         '%s.' % type(item))
                    prx = getattr(adp, method_name)(*args, **options)
                    prxs.append(prx)
                prx = DataSequenceTraversalProxy(prxs)
        return prx


@implementer(IDataTraversalProxyAdapter)
class DataTraversalProxyAdapter(object):
    """
    Abstract base class for data traversal proxy adapters.

    The adapters are used for dispatching a new data traversal proxy based
    on a resource data tree node's value.
    """
    #: The data traversal proxy class used by this factory.
    proxy_class = lambda *args: None

    def __init__(self, data):
        self._data = data

    def make_source_proxy(self, options=None):
        """
        Returns a data traversal proxy for the adapted data.
        """
        raise NotImplementedError('Abstract method.')

    def make_target_proxy(self, accessor, manage_back_references=True,
                          options=None):
        """
        Returns a data traversal proxy for the adapted data.
        """
        raise NotImplementedError('Abstract method.')

    def make_proxy(self, accessor, relationship_direction, options=None):
        """
        Returns a data traversal proxy for the adapted data.
        """
        if options is None:
            options = {}
        return self.proxy_class(self._data, accessor,
                                   relationship_direction, **options)


class ConvertingDataTraversalProxyMixin(object):
    """
    Mixin class for data traversal proxies that convert incoming
    representation data to an entity.
    """
    __converted_entity = None

    def get_entity(self):
        """
        Returns the entity converted from the proxied data.
        """
        if self._accessor is None:
            if self.__converted_entity is None:
                self.__converted_entity = self._convert_to_entity()
        else:
            # If we have an accessor, we can get the proxied entity by ID.
            # FIXME: This is a hack that is only used for REMOVE operations
            #        with data elements.
            self.__converted_entity = \
                self.get_matching(self.get_id()).get_entity()
        return self.__converted_entity

    def _convert_to_entity(self):
        raise NotImplementedError('Abstract method.')


class SourceTargetDataTreeTraverser(object):
    """
    Traverser for synchronous traversal of a source and a target data tree.

    For each data item pair, starting with the root of the source and target
    data tree, iterate over the non-terminal attributes of the associated type
    and obtain the attribute value (child data item) for both the source and
    the target. If the parent source or target data item is `None`, the
    corresponding child data item is also `None`.

    A child node is traversed along the ADD cascade when only the source
    was found, along the REMOVE cascade when only the target was found, and
    along the UPDATE cascade when both source and target were found. Traversal
    is suppressed when the parent attribute does not have the appropriate
    cascading flag set.

    When traversing along the ADD cascade, a child node of a node that is
    being added is only traversed (along the ADD cascade) when it has no ID
    (i.e., ID is None).

    When traversing along the REMOVE cascade, a child node of a node that is
    being removed is also removed if it has an ID (i.e., the "id" attribute
    is not None).
    """
    def __init__(self, source_proxy, target_proxy):
        self._src_prx = source_proxy
        self._tgt_prx = target_proxy
        self.__trv_path = TraversalPath()
        self.__root_is_sequence = \
            (not source_proxy is None and
             source_proxy.proxy_for == RESOURCE_KINDS.COLLECTION) \
            or (not target_proxy is None and
                target_proxy.proxy_for == RESOURCE_KINDS.COLLECTION)
        if __debug__:
            self.__logger = get_logger('everest.traversal')
        else:
            self.__logger = None

    @classmethod
    def make_traverser(cls, source_data, target_data, relation_operation,
                       accessor=None, manage_back_references=True,
                       source_proxy_options=None, target_proxy_options=None):
        """
        Factory method to create a tree traverser depending on the input
        source and target data combination.

        :param source_data: Source data.
        :param target_target: Target data.
        :param str relation_operation: Relation operation. On of the constants
          defined in :class:`everest.constants.RELATION_OPERATIONS`.
        :param accessor: Accessor for looking up target nodes for update
          operations.
        :param bool manage_back_references: Flag passed to the target proxy.
        """
        reg = get_current_registry()
        prx_fac = reg.getUtility(IDataTraversalProxyFactory)
        if relation_operation == RELATION_OPERATIONS.ADD \
           or relation_operation == RELATION_OPERATIONS.UPDATE:
            if relation_operation == RELATION_OPERATIONS.ADD \
               and not target_data is None:
                raise ValueError('Must not provide target data with '
                                 'relation operation ADD.')
            source_proxy = \
                    prx_fac.make_source_proxy(source_data,
                                              options=source_proxy_options)
            source_is_sequence = \
                source_proxy.proxy_for == RESOURCE_KINDS.COLLECTION
            if not source_is_sequence:
                source_id = source_proxy.get_id()
        else:
            source_proxy = None
            source_is_sequence = False
        if relation_operation == RELATION_OPERATIONS.REMOVE \
           or relation_operation == RELATION_OPERATIONS.UPDATE:
            if target_proxy_options is None:
                target_proxy_options = {}
            if relation_operation == RELATION_OPERATIONS.REMOVE:
                if not source_data is None:
                    raise ValueError('Must not provide source data with '
                                     'relation operation REMOVE.')
                target_proxy = prx_fac.make_target_proxy(
                                            target_data,
                                            accessor,
                                            manage_back_references=
                                                     manage_back_references,
                                            options=target_proxy_options)
            else: # UPDATE
                if accessor is None:
                    raise ValueError('Need to provide an accessor when '
                                     'performing UPDATE operations.')
                if not target_data is None:
                    target_root = target_data
                elif not source_is_sequence:
                    # Look up the (single) target to update.
                    target_root = accessor.get_by_id(source_id)
                    if target_root is None:
                        raise ValueError('Entity with ID %s to update not '
                                         'found.' % source_id)
                else:
                    # Look up collection of targets to update.
                    target_root = []
                    for src_prx in source_proxy:
                        tgt_ent_id = src_prx.get_id()
                        if tgt_ent_id is None:
                            continue
                        tgt_ent = accessor.get_by_id(tgt_ent_id)
                        if tgt_ent is None:
                            continue
                        target_root.append(tgt_ent)
                target_proxy = prx_fac.make_target_proxy(
                                            target_root,
                                            accessor,
                                            manage_back_references=
                                                 manage_back_references,
                                            options=target_proxy_options)
            target_is_sequence = \
                    target_proxy.proxy_for == RESOURCE_KINDS.COLLECTION
        else:
            target_proxy = None
            target_is_sequence = False
        if not source_proxy is None and not target_proxy is None:
            # Check for source/target consistency.
            if not ((source_is_sequence and target_is_sequence) or
                    (not source_is_sequence and not target_is_sequence)):
                raise ValueError('When both source and target root nodes are '
                                 'given, they can either both be sequences '
                                 'or both not be sequences.')
        return cls(source_proxy, target_proxy)

    def run(self, visitor):
        """
        :param visitor: visitor to call with every node in the domain tree.
        :type visitor: subclass of
            :class:`everest.entities.traversal.DomainVisitor`
        """
        if __debug__:
            self.__log_run(visitor)
        visitor.prepare()
        if self.__root_is_sequence:
            if not self._tgt_prx is None:
                tgts = iter(self._tgt_prx)
            else:
                tgts = None
            if not self._src_prx is None:
                srcs = iter(self._src_prx)
            else:
                srcs = None
            self.traverse_many(None, srcs, tgts, visitor)
        else:
            self.traverse_one(None, self._src_prx, self._tgt_prx, visitor)
        visitor.finalize()

    def traverse_one(self, attribute, source, target, visitor):
        """
        :param source: source data proxy
        :type source: instance of `DataTraversalProxy` or None
        :param target: target data proxy
        :type target: instance of `DataTraversalProxy` or None
        """
        if __debug__:
            self.__log_traverse_one(self.__trv_path, attribute, source, target)
        prx = source or target
        if prx.do_traverse():
            rel_op = RELATION_OPERATIONS.check(source, target)
            for attr in prx.get_relationship_attributes():
                # Check cascade settings.
                if not bool(attr.cascade & rel_op):
                    continue
                if not source is None:
                    try:
                        attr_source = source.get_attribute_proxy(attr)
                    except AttributeError:
                        # If the source does not have the attribute set, we
                        # do nothing (as opposed to when the value is None).
                        continue
                else:
                    attr_source = None
                if not target is None:
                    attr_target = target.get_attribute_proxy(attr)
                else:
                    attr_target = None
                attr_rel_op = RELATION_OPERATIONS.check(attr_source,
                                                        attr_target)
                if attr_rel_op == RELATION_OPERATIONS.ADD:
                    if rel_op == RELATION_OPERATIONS.ADD:
                        parent = source
                    else:
                        parent = target
                elif attr_rel_op == RELATION_OPERATIONS.REMOVE:
                    parent = target
                else: # UPDATE
                    parent = target
                card = get_attribute_cardinality(attr)
                if card == CARDINALITY_CONSTANTS.ONE:
                    if attr_source is None and attr_target is None:
                        # If both source and target have None values, there is
                        # nothing to do.
                        continue
                    if attr_rel_op == RELATION_OPERATIONS.ADD:
#                        if not attr_source.get_id() is None:
#                            # We only ADD new items.
#                            continue
                        src_items = [attr_source]
                        tgt_items = None
                    elif attr_rel_op == RELATION_OPERATIONS.REMOVE:
                        src_items = None
                        tgt_items = [attr_target]
                    else: # UPDATE
                        src_items = [attr_source]
                        tgt_items = [attr_target]
                        src_id = attr_source.get_id()
                        tgt_id = attr_target.get_id()
                        if src_id != tgt_id:
                            if not src_id is None:
                                # If the source ID is None, this is a replace
                                # operation (ADD source, REMOVE target).
                                src_target = attr_target.get_matching(src_id)
                                if not src_target is None:
                                    tgt_items.append(src_target)
                else:
                    src_items = attr_source
                    tgt_items = attr_target
                self.__trv_path.push(parent, (source, target), attr, rel_op)
                self.traverse_many(attr, src_items, tgt_items, visitor)
                self.__trv_path.pop() # path.pop()
        visitor.visit(self.__trv_path, attribute, source, target)

    def traverse_many(self, attribute, source_sequence, target_sequence,
                      visitor):
        """
        Traverses the given source and target sequences and makes appropriate
        calls to :method:`traverse_one`.

        Algorithm:
        1) Build a map target item ID -> target data item from the target
           sequence;
        2) For each source data item in the source sequence check if it
           has a not-None ID; if yes, remove the corresponding target from the
           map generated in step 1) and use as target data item for the
           source data item; if no, use `None` as target data item;
        3) For the remaining items in the target map from 1), call
           :method:`traverse_one` passing `None` as source (REMOVE);
        4) For all source/target data item pairs generated in 2, call
           :method:`traverse_one` (ADD or UPDATE depending on whether target
           item is `None`).

        :param source_sequence: iterable of source data proxies
        :type source_sequence: iterator yielding instances of
                               `DataTraversalProxy` or None
        :param target_sequence: iterable of target data proxies
        :type target_sequence: iterator yielding instances of
                               `DataTraversalProxy` or None
        """
        target_map = {}
        if not target_sequence is None:
            for target in target_sequence:
                target_map[target.get_id()] = target
        src_tgt_pairs = []
        if not source_sequence is None:
            for source in source_sequence:
                source_id = source.get_id()
                if not source_id is None:
                    # Check if target exists for UPDATE.
                    target = target_map.pop(source_id, None)
                else:
                    # Source is new, there is no target, so ADD.
                    target = None
                src_tgt_pairs.append((source, target))
        # All targets that are now still in the map where not present in the
        # source and therefore need to be REMOVEd.
        for target in itervalues_(target_map):
            if not (None, target) in self.__trv_path:
                self.traverse_one(attribute, None, target, visitor)
        #
        for source, target in src_tgt_pairs:
            if not (source, target) in self.__trv_path:
                self.traverse_one(attribute, source, target, visitor)

    def __log_run(self, visitor):
        self.__logger.debug('Traversing %s->%s with %s'
                            % (self._src_prx, self._tgt_prx, visitor))

    def __log_traverse_one(self, path, attribute, source, target):
        if target is None:
            mode = 'ADD'
            data = '%s,None' % source
        elif source is None:
            mode = 'REMOVE'
            data = 'None,%s' % target
        else:
            mode = 'UPDATE'
            data = '%s,%s' % (source, target)
        if not attribute is None:
            parent = "(%s)" % path.parent
            self.__logger.debug('%s%s %s.%s (%s)' %
                                ("  "*len(path), mode, parent,
                                 attribute.resource_attr, data))
        else:
            self.__logger.debug('%s ROOT (%s)' % (mode, data))



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 10, 2011.
"""
from sqlalchemy.engine import create_engine
from sqlalchemy.orm.scoping import scoped_session
from sqlalchemy.orm.session import sessionmaker

from everest.querying.filtering import CqlFilterSpecificationVisitor
from everest.querying.ordering import CqlOrderSpecificationVisitor
from everest.querying.specifications import FilterSpecificationFactory
from everest.querying.specifications import OrderSpecificationFactory
from everest.repositories.rdb import SqlFilterSpecificationVisitor
from everest.repositories.rdb import SqlOrderSpecificationVisitor
from everest.repositories.rdb.querying import OrderClauseList
from everest.repositories.rdb.utils import reset_metadata
from everest.testing import Pep8CompliantTestCase
import sqlalchemy as sa
import sqlalchemy.orm as orm


__docformat__ = 'reStructuredText en'
__all__ = ['CqlFilterSpecificationVisitorTestCase',
           'CqlOrderSpecificationVisitorTestCase',
           'SqlFilterSpecificationVisitorTestCase',
           'SqlOrderSpecificationVisitorTestCase',
           ]


class Person(object):
    metadata = None
    id = None
    name = None
    age = None

    def __init__(self, name, age):
        self.name = name
        self.age = age


def create_metadata(engine):
    metadata = sa.MetaData()
    person_table = sa.Table('person', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('name', sa.String, nullable=False, unique=True),
        sa.Column('age', sa.Integer, nullable=False),
        )
    orm.mapper(Person, person_table)
    metadata.bind = engine
    metadata.create_all()
    return metadata


def teardown():
    # Module level tear down.
    if not Person.metadata is None:
        Person.metadata.drop_all()
        Person.metadata = None
    reset_metadata()


class VisitorTestCase(Pep8CompliantTestCase):
    visitor = None
    specs_factory = None
    _spec_map = None

    def set_up(self):
        Pep8CompliantTestCase.set_up(self)
        self.visitor = self._make_visitor()
        self.specs_factory = self._make_specs_factory()

    def _get_spec(self, name):
        return self._get_spec_map()[name]

    def _get_spec_map(self):
        if self._spec_map is None:
            self._spec_map = self._init_spec_map()
        return self._spec_map

    def _run_visitor(self, spec_name):
        spec = self._get_spec(spec_name)
        spec.accept(self.visitor)
        return self.visitor.expression

    def _make_visitor(self):
        raise NotImplementedError('Abstract method.')

    def _make_specs_factory(self):
        raise NotImplementedError('Abstract method.')

    def _init_spec_map(self):
        raise NotImplementedError('Abstract method.')


class FilterVisitorTestCase(VisitorTestCase):

    def _make_visitor(self):
        raise NotImplementedError('Abstract method.')

    def _make_specs_factory(self):
        return FilterSpecificationFactory()

    def _init_spec_map(self):
        sm = {}
        sm['starts-with'] = \
            self.specs_factory.create_starts_with('name', 'Ni')
        sm['ends-with'] = \
            self.specs_factory.create_ends_with('name', 'os')
        sm['contains'] = \
            self.specs_factory.create_contains('name', 'iko')
        sm['contained'] = \
            self.specs_factory.create_contained('age', [22, 33, 44, 55])
        sm['equal-to'] = \
            self.specs_factory.create_equal_to('name', 'Nikos')
        sm['less-than'] = \
            self.specs_factory.create_less_than('age', 34)
        sm['less-than-or-equal-to'] = \
            self.specs_factory.create_less_than_or_equal_to('age', 34)
        sm['greater-than'] = \
            self.specs_factory.create_greater_than('age', 34)
        sm['greater-than-or-equal-to'] = \
            self.specs_factory.create_greater_than_or_equal_to('age', 34)
        sm['in-range'] = \
            self.specs_factory.create_in_range('age', (30, 40))
        left_spec = self.specs_factory.create_greater_than('age', 34)
        right_spec = self.specs_factory.create_equal_to('name', 'Nikos')
        sm['conjunction'] = left_spec & right_spec
        spec1 = self.specs_factory.create_equal_to('age', 34)
        spec2 = self.specs_factory.create_equal_to('age', 44)
        sm['disjunction'] = spec1 | spec2
        spec_a1 = self.specs_factory.create_equal_to('age', 34)
        spec_a2 = self.specs_factory.create_equal_to('age', 44)
        spec_a = spec_a1 | spec_a2
        spec_b1 = self.specs_factory.create_equal_to('name', 'Nikos')
        spec_b2 = self.specs_factory.create_equal_to('name', 'Oliver')
        spec_b = spec_b1 | spec_b2
        sm['conjunction-with-disjunction'] = spec_a & spec_b
        sm['not-starts-with'] = ~sm['starts-with']
        sm['not-ends-with'] = ~sm['ends-with']
        sm['not-contains'] = ~sm['contains']
        sm['not-contained'] = ~sm['contained']
        sm['not-equal-to'] = ~sm['equal-to']
        sm['not-less-than'] = ~sm['less-than']
        sm['not-less-than-or-equal-to'] = ~sm['less-than-or-equal-to']
        sm['not-greater-than'] = ~sm['greater-than']
        sm['not-greater-than-or-equal-to'] = ~sm['greater-than-or-equal-to']
        sm['not-in-range'] = ~sm['in-range']
        return sm


class CqlFilterSpecificationVisitorTestCase(FilterVisitorTestCase):

    def _make_visitor(self):
        return CqlFilterSpecificationVisitor()

    def test_visit_value_starts_with(self):
        expected_cql = 'name:starts-with:"Ni"'
        expr = self._run_visitor('starts-with')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_ends_with(self):
        expected_cql = 'name:ends-with:"os"'
        expr = self._run_visitor('ends-with')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_contains(self):
        expected_cql = 'name:contains:"iko"'
        expr = self._run_visitor('contains')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_contained(self):
        expected_cql = 'age:contained:22,33,44,55'
        expr = self._run_visitor('contained')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_equal_to(self):
        expected_cql = 'name:equal-to:"Nikos"'
        expr = self._run_visitor('equal-to')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_less_than(self):
        expected_cql = 'age:less-than:34'
        expr = self._run_visitor('less-than')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_less_or_equals(self):
        expected_cql = 'age:less-than-or-equal-to:34'
        expr = self._run_visitor('less-than-or-equal-to')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_greater_than(self):
        expected_cql = 'age:greater-than:34'
        expr = self._run_visitor('greater-than')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_greater_or_equals(self):
        expected_cql = 'age:greater-than-or-equal-to:34'
        expr = self._run_visitor('greater-than-or-equal-to')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_in_range(self):
        expected_cql = 'age:in-range:30-40'
        expr = self._run_visitor('in-range')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_Conjunction(self):
        expected_cql = 'age:greater-than:34~name:equal-to:"Nikos"'
        expr = self._run_visitor('conjunction')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_disjuction(self):
        expected_cql = 'age:equal-to:34,44'
        expr = self._run_visitor('disjunction')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_Conjunction_with_disjuction(self):
        expected_cql = 'age:equal-to:34,44~name:equal-to:"Nikos","Oliver"'
        expr = self._run_visitor('conjunction-with-disjunction')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_starts_with(self):
        expected_cql = 'name:not-starts-with:"Ni"'
        expr = self._run_visitor('not-starts-with')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_ends_with(self):
        expected_cql = 'name:not-ends-with:"os"'
        expr = self._run_visitor('not-ends-with')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_contains(self):
        expected_cql = 'name:not-contains:"iko"'
        expr = self._run_visitor('not-contains')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_equal_to_many(self):
        expected_cql = 'age:not-contained:22,33,44,55'
        expr = self._run_visitor('not-contained')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_equal_to(self):
        expected_cql = 'name:not-equal-to:"Nikos"'
        expr = self._run_visitor('not-equal-to')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_less_than(self):
        expected_cql = 'age:greater-than-or-equal-to:34'
        expr = self._run_visitor('not-less-than')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_less_or_equals(self):
        expected_cql = 'age:greater-than:34'
        expr = self._run_visitor('not-less-than-or-equal-to')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_greater_than(self):
        expected_cql = 'age:less-than-or-equal-to:34'
        expr = self._run_visitor('not-greater-than')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_greater_or_equals(self):
        expected_cql = 'age:less-than:34'
        expr = self._run_visitor('not-greater-than-or-equal-to')
        self.assert_equal(str(expr), expected_cql)

    def test_visit_value_not_in_range(self):
        expected_cql = 'age:not-in-range:30-40'
        expr = self._run_visitor('not-in-range')
        self.assert_equal(str(expr), expected_cql)


class SqlFilterSpecificationVisitorTestCase(FilterVisitorTestCase):
    def set_up(self):
        if Person.metadata is None:
            reset_metadata()
            engine = create_engine('sqlite://')
            metadata = create_metadata(engine)
            Person.metadata = metadata
            metadata.bind = engine
        VisitorTestCase.set_up(self)

    def _make_visitor(self):
        return SqlFilterSpecificationVisitor(Person)

    def test_visit_value_starts_with(self):
        expected_expr = Person.name.startswith('Ni')
        expr = self._run_visitor('starts-with')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_ends_with(self):
        expected_expr = Person.name.endswith('os')
        expr = self._run_visitor('ends-with')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_contains(self):
        expected_expr = Person.name.contains('iko')
        expr = self._run_visitor('contains')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_contained(self):
        expected_expr = Person.age.in_([22, 33, 44, 55])
        expr = self._run_visitor('contained')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_equal_to(self):
        expected_expr = Person.name == 'Nikos'
        expr = self._run_visitor('equal-to')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_less_than(self):
        expected_expr = Person.age < 34
        expr = self._run_visitor('less-than')
        self.assert_equal(str(expr), str(expected_expr))

    def _test_visit_value_less_or_equals(self):
        expected_expr = Person.age <= 34
        expr = self._run_visitor('less-than-or-equal-to')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_greater_than(self):
        expected_expr = Person.age > 34
        expr = self._run_visitor('greater-than')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_greater_or_equals(self):
        expected_expr = Person.age >= 34
        expr = self._run_visitor('greater-than-or-equal-to')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_in_range(self):
        expected_expr = Person.age.between(30, 40)
        expr = self._run_visitor('in-range')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_Conjunction(self):
        expected_expr = sa.and_(Person.age > 34, Person.name == 'Nikos')
        expr = self._run_visitor('conjunction')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_disjuction(self):
        expected_expr = sa.or_(Person.age == 34, Person.age == 44)
        expr = self._run_visitor('disjunction')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_Conjunction_with_list_equality(self):
        expected_expr = sa.and_(Person.age.in_([34, 44]),
                                Person.name.in_(['Nikos', 'Oliver']))
        spec_a = self.specs_factory.create_contained('age', [34, 44])
        spec_b = self.specs_factory.create_contained('name',
                                                     ['Nikos', 'Oliver'])
        spec = spec_a & spec_b
        spec.accept(self.visitor)
        self.assert_equal(str(self.visitor.expression),
                          str(expected_expr))

    def test_visit_value_not_starts_with(self):
        expected_expr = sa.not_(Person.name.startswith('Ni'))
        expr = self._run_visitor('not-starts-with')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_ends_with(self):
        expected_expr = sa.not_(Person.name.endswith('os'))
        expr = self._run_visitor('not-ends-with')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_contains(self):
        expected_expr = sa.not_(Person.name.contains('iko'))
        expr = self._run_visitor('not-contains')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_equal_to(self):
        expected_expr = sa.not_(Person.name == 'Nikos')
        expr = self._run_visitor('not-equal-to')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_less_than(self):
        expected_expr = sa.not_(Person.age < 34)
        expr = self._run_visitor('not-less-than')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_less_or_equals(self):
        expected_expr = sa.not_(Person.age <= 34)
        expr = self._run_visitor('not-less-than-or-equal-to')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_greater_than(self):
        expected_expr = sa.not_(Person.age > 34)
        expr = self._run_visitor('not-greater-than')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_greater_or_equals(self):
        expected_expr = sa.not_(Person.age >= 34)
        expr = self._run_visitor('not-greater-than-or-equal-to')
        self.assert_equal(str(expr), str(expected_expr))

    def test_visit_value_not_in_range(self):
        expected_expr = sa.not_(Person.age.between(30, 40))
        expr = self._run_visitor('not-in-range')
        self.assert_equal(str(expr), str(expected_expr))


class OrderVisitorTestCase(VisitorTestCase):

    def _make_visitor(self):
        raise NotImplementedError('Abstract method.')

    def _make_specs_factory(self):
        return OrderSpecificationFactory()

    def _init_spec_map(self):
        sm = {}
        sm['one-asc'] = self.specs_factory.create_ascending('name')
        sm['one-desc'] = self.specs_factory.create_descending('name')
        sm['two-asc-asc'] = self.specs_factory.create_ascending('name') \
                            & self.specs_factory.create_ascending('age')
        sm['two-desc-asc'] = self.specs_factory.create_descending('name') \
                             & self.specs_factory.create_ascending('age')
        sm['two-asc-desc'] = self.specs_factory.create_ascending('name') \
                             & self.specs_factory.create_descending('age')
        sm['two-desc-desc'] = self.specs_factory.create_descending('name') \
                              & self.specs_factory.create_descending('age')
        return sm


class SqlOrderSpecificationVisitorTestCase(OrderVisitorTestCase):

    def _make_visitor(self):
        return SqlOrderSpecificationVisitor(Person)

    def set_up(self):
        if Person.metadata is None:
            engine = create_engine('sqlite://')
            Person.metadata = create_metadata(engine)
        OrderVisitorTestCase.set_up(self)

    def test_simple_order_by_one_attribute(self):
        expected_expr = Person.name.asc()
        expr = self._run_visitor('one-asc')
        self.assert_equal(str(expr), str(expected_expr))

    def test_simple_reversed_order_by_one_attribute(self):
        expected_expr = Person.name.desc()
        expr = self._run_visitor('one-desc')
        self.assert_equal(str(expr), str(expected_expr))

    def test_simple_order_by_two_attributes(self):
        expected_expr = OrderClauseList(Person.name.asc(), Person.age.asc())
        expr = self._run_visitor('two-asc-asc')
        self.assert_equal(str(expr), str(expected_expr))

    def test_simple_order_by_two_attributes_left_reversed(self):
        expected_expr = OrderClauseList(Person.name.desc(), Person.age.asc())
        expr = self._run_visitor('two-desc-asc')
        self.assert_equal(str(expr), str(expected_expr))

    def test_simple_order_by_two_attributes_right_reversed(self):
        expected_expr = OrderClauseList(Person.name.asc(), Person.age.desc())
        expr = self._run_visitor('two-asc-desc')
        self.assert_equal(str(expr), str(expected_expr))

    def test_simple_order_by_two_attributes_both_reversed(self):
        expected_expr = OrderClauseList(Person.name.desc(), Person.age.desc())
        expr = self._run_visitor('two-desc-desc')
        self.assert_equal(str(expr), str(expected_expr))

    def test_order_clause_list(self):
        # This emulates customized comparators which return clause lists
        # for .asc and .desc operations.
        old_asc = Person.name.asc
        expected_expr = OrderClauseList(Person.id.asc(),
                                        Person.name.asc(),
                                        Person.age.asc())
        self.assert_equal(str(expected_expr),
                          'person.id ASC, person.name ASC, person.age ASC')
        try:
            Person.name.asc = lambda : OrderClauseList(Person.id.asc(),
                                                       old_asc())
            expr = self._run_visitor('two-asc-asc')
            self.assert_equal(str(expr), str(expected_expr))
        finally:
            Person.name.asc = old_asc
        # Make sure the correct ORDER BY clause is generated.
        sm = scoped_session(sessionmaker())
        sess = sm()
        q = sess.query(Person).order_by(expr) # pylint: disable=E1101
        q_str = str(q.statement)
        self.assert_not_equal(q_str.find("ORDER BY %s" % expr), -1)


class CqlOrderSpecificationVisitorTestCase(OrderVisitorTestCase):
    visitor = None
    order_factory = None

    def _make_visitor(self):
        return CqlOrderSpecificationVisitor()

    def test_simple_order_by_one_attribute(self):
        expected_cql = 'name:asc'
        expr = self._run_visitor('one-asc')
        self.assert_equal(str(expr), expected_cql)

    def test_simple_reversed_order_by_one_attribute(self):
        expected_cql = 'name:desc'
        expr = self._run_visitor('one-desc')
        self.assert_equal(str(expr), expected_cql)

    def test_simple_order_by_two_attributes(self):
        expected_cql = 'name:asc~age:asc'
        expr = self._run_visitor('two-asc-asc')
        self.assert_equal(str(expr), expected_cql)

    def test_simple_order_by_two_attributes_left_reversed(self):
        expected_cql = 'name:desc~age:asc'
        expr = self._run_visitor('two-desc-asc')
        self.assert_equal(str(expr), expected_cql)

    def test_simple_order_by_two_attributes_right_reversed(self):
        expected_cql = 'name:asc~age:desc'
        expr = self._run_visitor('two-asc-desc')
        self.assert_equal(str(expr), expected_cql)

    def test_simple_order_by_two_attributes_both_reversed(self):
        expected_cql = 'name:desc~age:desc'
        expr = self._run_visitor('two-desc-desc')
        self.assert_equal(str(expr), expected_cql)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 7, 2012.
"""
from everest.mime import CsvMime
from everest.representers.attributes import MappedAttributeKey
from everest.representers.config import IGNORE_OPTION
from everest.representers.interfaces import IDataElement
from everest.representers.interfaces import IRepresenterRegistry
from everest.representers.traversal import DataElementTreeTraverser
from everest.representers.utils import NewRepresenterConfigurationContext
from everest.representers.utils import UpdatedRepresenterConfigurationContext
from everest.testing import ResourceTestCase
from everest.tests.complete_app.resources import MyEntityMember
from zope.interface import alsoProvides as also_provides # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['MappingTestCase',
           ]


class MappingTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_rpr.zcml'

    def set_up(self):
        ResourceTestCase.set_up(self)
        rpr_reg = self.config.get_registered_utility(IRepresenterRegistry)
        self.mapping_registry = rpr_reg.get_mapping_registry(CsvMime)
        self.mapping = self.mapping_registry.find_mapping(MyEntityMember)

    def test_attribute_iterators(self):
        self.assert_true(self.mapping.mapped_class is MyEntityMember)
        attrs = list(self.mapping.attribute_iterator())
        t_attrs = list(self.mapping.terminal_attribute_iterator())
        nt_attrs = list(self.mapping.nonterminal_attribute_iterator())
        self.assert_equal(len(attrs), len(t_attrs) + len(nt_attrs))

    def test_register_non_resource_raises_error(self):
        self.assert_raises(ValueError, self.mapping_registry.create_mapping,
                           NonResource)

    def test_mapping_non_resource_raises_error(self):
        non_rc = NonResource()
        self.assert_raises(ValueError, self.mapping.map_to_data_element,
                           non_rc)

    def test_mapping_non_dataelement_raises_error(self):
        non_de = NonDataElement()
        self.assert_raises(ValueError, self.mapping.map_to_resource,
                           non_de)

    def test_mapping_invalid_dataelement_raises_error(self):
        invalid_de = InvalidKindDataElement()
        invalid_de.mapping = self.mapping # pylint: disable=W0201
        also_provides(invalid_de, IDataElement)
        trv = DataElementTreeTraverser(invalid_de, self.mapping.as_pruning())
        self.assert_raises(ValueError, trv.run, None)

    def test_mapping_access(self):
        key = MappedAttributeKey(())
        self.assert_true(str(key).startswith(key.__class__.__name__))
        attr_map = self.mapping.get_attribute_map(key=key)
        self.assert_true(attr_map['children'].options[IGNORE_OPTION]
                         is False)

    def test_representer_configuration_contexts(self):
        self.assert_raises(IndexError, self.mapping.pop_configuration)
        opts = dict(parent={IGNORE_OPTION : True})
        self.assert_false(
            self.mapping.configuration.get_attribute_option('parent',
                                                            IGNORE_OPTION))
        self.assert_false(
            self.mapping.configuration.get_attribute_option('children',
                                                            IGNORE_OPTION))
        ctx1 = NewRepresenterConfigurationContext(MyEntityMember, CsvMime,
                                                  attribute_options=opts)
        with ctx1:
            mp1 = self.mapping_registry.find_mapping(MyEntityMember)
            self.assert_true(
                    mp1.configuration.get_attribute_option('parent',
                                                           IGNORE_OPTION))
            # Changed to default.
            self.assert_is_none(
                    mp1.configuration.get_attribute_option('children',
                                                           IGNORE_OPTION))
        ctx2 = UpdatedRepresenterConfigurationContext(MyEntityMember, CsvMime,
                                                      attribute_options=opts)
        with ctx2:
            mp2 = self.mapping_registry.find_mapping(MyEntityMember)
            self.assert_true(
                    mp2.configuration.get_attribute_option('parent',
                                                           IGNORE_OPTION))
            # Unchanged.
            self.assert_false(
                    mp2.configuration.get_attribute_option('children',
                                                           IGNORE_OPTION))

class NonResource(object):
    pass


class NonDataElement(object):
    pass


class InvalidKindDataElement(object):
    def __init__(self):
        self.kind = None



"""
Filter CQL criteria expression parser.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 5, 2011.
"""
from iso8601.iso8601 import ParseError
from iso8601.iso8601 import parse_date
from pyparsing import CaselessKeyword
from pyparsing import Combine
from pyparsing import Empty
from pyparsing import Forward
from pyparsing import Group
from pyparsing import Literal
from pyparsing import OneOrMore
from pyparsing import Optional
from pyparsing import Regex
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import alphanums
from pyparsing import alphas
from pyparsing import dblQuotedString
from pyparsing import delimitedList
from pyparsing import nums
from pyparsing import opAssoc
from pyparsing import operatorPrecedence
from pyparsing import replaceWith
from pyparsing import sglQuotedString
from pyparsing import srange
from pyramid.compat import string_types

from everest.entities.utils import identifier_from_slug
from everest.querying.operators import CONTAINED
from everest.querying.operators import CONTAINS
from everest.querying.operators import ENDS_WITH
from everest.querying.operators import EQUAL_TO
from everest.querying.operators import GREATER_OR_EQUALS
from everest.querying.operators import GREATER_THAN
from everest.querying.operators import IN_RANGE
from everest.querying.operators import LESS_OR_EQUALS
from everest.querying.operators import LESS_THAN
from everest.querying.operators import STARTS_WITH
from everest.querying.specifications import cntd
from everest.querying.specifications import cnts
from everest.querying.specifications import ends
from everest.querying.specifications import eq
from everest.querying.specifications import ge
from everest.querying.specifications import gt
from everest.querying.specifications import le
from everest.querying.specifications import lt
from everest.querying.specifications import rng
from everest.querying.specifications import starts
from everest.resources.interfaces import ICollectionResource
from everest.resources.utils import url_to_resource


__docformat__ = 'reStructuredText en'
__all__ = ['parse_filter',
           ]

# Adapted from the iso8601 package to *require* a full yyyy-mm-ddThh:mm:ssZ
# string at the minimum.
ISO8601_REGEX = r'(?P<year>[0-9]{4})' \
                r'-(?P<month>[0-9]{1,2})' \
                r'-(?P<day>[0-9]{1,2})' \
                r'(?P<separator>.)' \
                r'(?P<hour>[0-9]{2})' \
                r':(?P<minute>[0-9]{2})' \
                r':(?P<second>[0-9]{2})' \
                r'(\\.(?P<fraction>[0-9]+))?' \
                r'(?P<timezone>Z|(([-+])([0-9]{2}):([0-9]{2})))?'


AND_PAT = 'and'
OR_PAT = 'or'

BINARY = 2

colon = Literal(':')
comma = Literal(',')
dot = Literal('.')
tilde = Literal('~')
slash = Literal('/')
open_paren = Literal('(')
close_paren = Literal(')')
dbl_quote = Literal('"')
nonzero_nums = srange('[1-9]')
empty = Empty()
true = CaselessKeyword("true").setParseAction(replaceWith(True))
false = CaselessKeyword("false").setParseAction(replaceWith(False))
attribute = Word(alphas, alphanums + '-')
identifier = \
    Combine(attribute + ZeroOrMore('.' + attribute)).setName('identifier')
and_op = CaselessKeyword(AND_PAT).setParseAction(replaceWith(AND_PAT))
or_op = CaselessKeyword(OR_PAT).setParseAction(replaceWith(OR_PAT))


def convert_number(toks):
    str_val = toks[0]
    if '.' in str_val or 'e' in str_val:
        val = float(str_val)
    else:
        val = int(str_val)
    return val


def convert_date(toks):
    date_val = toks[0][0]
    try:
        res = parse_date(date_val)
    except ParseError:
        res = date_val
    return res


def convert_range(toks):
    return (toks.range[0], toks.range[-1])


class CriterionConverter(object):
    spec_map = {STARTS_WITH.name : starts,
                ENDS_WITH.name : ends,
                CONTAINS.name : cnts,
                CONTAINED.name : cntd,
                EQUAL_TO.name : eq,
                GREATER_THAN.name : gt,
                LESS_THAN.name : lt,
                GREATER_OR_EQUALS.name : ge,
                LESS_OR_EQUALS.name : le,
                IN_RANGE.name : rng,
                }

    @classmethod
    def convert(cls, toks):
        crit = toks[0]
        # Extract attribute name.
        attr_name = cls.__prepare_identifier(crit.name)
        # Extract operator name.
        op_name = cls.__prepare_identifier(crit.operator)
        if op_name.startswith("not_"):
            op_name = op_name[4:]
            negate = True
        else:
            negate = False
        # Extract attribute value.
        if len(crit.value) == 0:
            raise ValueError('Criterion does not define a value.')
#        elif len(crit.value) == 1 and isinstance(crit.value, ParseResults):
#            # URLs - convert to resource.
#            url_val = crit.value
#            try:
#                rc = url_to_resource(url_val.resource)
#            except:
#                raise ValueError('Could not convert "%s" to a resource.'
#                                 % url_val.resource)
#            if not url_val.query == '':
#                if not ICollectionResource.providedBy(rc): # pylint: disable=E1101
#                    raise ValueError('Member resources can not have a '
#                                     'query string.')
#                rc.filter = url_val.query
#            attr_value = rc
#            value_is_resource = True
        elif len(crit.value) == 1 \
             and ICollectionResource.providedBy(crit.value[0]): # pylint: disable=E1101
            attr_value = crit.value[0]
            value_is_resource = True
        else:
            attr_value = cls.__prepare_values(crit.value)
            value_is_resource = False
        spec_gen = cls.spec_map[op_name]
        if op_name == CONTAINED.name or value_is_resource:
            spec = spec_gen(**{attr_name:attr_value})
            if negate:
                spec = ~spec
        else:
            # Create a spec for each value and concatenate with OR.
            spec = cls.__make_spec(spec_gen, attr_name, attr_value, negate)
        return spec

    @classmethod
    def __make_spec(cls, spec_gen, attr_name, attr_values, negate):
        spec = None
        for attr_value in attr_values:
            cur_spec = spec_gen(**{attr_name:attr_value})
            if negate:
                cur_spec = ~cur_spec
            if spec is None:
                spec = cur_spec
            else:
                spec = spec | cur_spec
        return spec

    @classmethod
    def __prepare_identifier(cls, name):
        return identifier_from_slug(name)

    @classmethod
    def __prepare_values(cls, values):
        prepared = []
        for val in values:
            if not cls.__is_empty_string(val) and not val in prepared:
                prepared.append(val)
        return prepared

    @classmethod
    def __is_empty_string(cls, v):
        return isinstance(v, string_types) and len(v) == 0


def convert_conjunction(toks):
    left_spec = toks[0][0]
    right_spec = toks[0][-1]
    return left_spec & right_spec


def convert_disjunction(toks):
    left_spec = toks[0][0]
    right_spec = toks[0][-1]
    return left_spec | right_spec


def convert_simple_criteria(toks):
    spec = None
    for crit_spec in toks[0]:
        if spec is None:
            spec = crit_spec
        else:
            spec = spec & crit_spec
    return spec



def convert_string(toks):
    unquoted = toks[0][1:-1]
    if len(url_protocol.searchString(unquoted)) > 0:
        result = [url_to_resource(unquoted)]
    else:
        result = [unquoted]
    return result


# Numbers are converted to ints if possible.
cql_number = Combine(Optional('-') + ('0' | Word(nonzero_nums, nums)) +
                     Optional('.' + Word(nums)) +
                     Optional(Word('eE', exact=1) + Word(nums + '+-', nums))
                     ).setParseAction(convert_number)
# Dates are parsed as double-quoted ISO8601 strings and converted to datetime
# objects.
cql_date = Combine(dbl_quote.suppress() + Regex(ISO8601_REGEX) +
                   dbl_quote.suppress()
                   ).setParseAction(convert_date)
# All double-quoted strings that are not dates are returned with their quotes
# removed.
cql_string = (dblQuotedString | sglQuotedString).setParseAction(convert_string)

# URLs are detected as strings starting with the http(s) protocol.
url_protocol = Combine(Literal('http') + Optional('s'))

# Number range.
# FIXME: char ranges are not supported yet
cql_number_range = Group(cql_number + '-' + cql_number
                           ).setParseAction(convert_range)

cql_values = Group(delimitedList(
                        cql_number_range('range') |
                        cql_number('number') |
                        cql_date('date') |
                        cql_string('string') |
                        true |
                        false |
                        empty
                        )
                   )

logical_op = and_op('operator') | or_op('operator')

criterion = Group(identifier('name') + colon.suppress() +
                  identifier('operator') + colon.suppress() +
                  cql_values('value')
                  )
criterion.setParseAction(CriterionConverter.convert)

# Recursive definition of "junction" (AND or OR) clauses.
junctions = Forward()
junction_element = (criterion |
                    open_paren.suppress() + junctions +
                    close_paren.suppress())
junctions << operatorPrecedence(# pylint: disable=W0106
                    junction_element,
                    [(and_op, BINARY, opAssoc.LEFT, convert_conjunction),
                     (or_op, BINARY, opAssoc.LEFT, convert_disjunction),
                     ])

# Old-style criteria separated by "~".
simple_criteria = Group(criterion +
                        OneOrMore(tilde.suppress() + criterion))
simple_criteria.setParseAction(convert_simple_criteria)

query = (simple_criteria | junctions) # pylint: disable=W0104


def parse_filter(query_string):
    """
    Parses the given filter criteria string.
    """
    return query.parseString(query_string)[0]



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
from pyramid.compat import urlparse
import pytest

from everest.resources.utils import resource_to_url
from everest.resources.utils import url_to_resource
from everest.tests.complete_app.interfaces import IMyEntityParent
from everest.tests.complete_app.resources import MyEntityMember
from everest.resources.service import Service


__docformat__ = 'reStructuredText en'
__all__ = ['TestUrlNoRdb',
           'TestUrlRdb',
           ]


@pytest.mark.usefixtures("collection")
class _TestUrl(object):
    package_name = 'everest.tests.complete_app'
    app_url = 'http://0.0.0.0:6543'
    base_url = '%s/my-entities/' % app_url

    def test_resource_to_url_non_resource_object(self, member):
        ent = member.get_entity()
        with pytest.raises(TypeError) as cm:
            resource_to_url(ent)
        exc_msg = 'Can not generate URL for non-resource'
        assert str(cm.value).startswith(exc_msg)

    def test_resource_to_url_floating_member(self, member):
        ent = member.get_entity()
        mb = MyEntityMember.create_from_entity(ent)
        with pytest.raises(ValueError) as cm:
            resource_to_url(mb)
        exc_msg = 'Can not generate URL for floating resource'
        assert str(cm.value).startswith(exc_msg)

    def test_resource_to_url_member(self, member):
        self.__check_url(resource_to_url(member),
                         schema='http', path='/my-entities/0/', params='',
                         query='')

    def test_resource_to_url_collection(self, collection):
        self.__check_url(resource_to_url(collection),
                         schema='http', path='/my-entities/', params='',
                         query='')

    def test_resource_to_url_with_slice(self, collection):
        collection.slice = slice(0, 1)
        self.__check_url(resource_to_url(collection),
                         schema='http', path='/my-entities/',
                         params='', query='start=0&size=1')

    def test_resource_to_url_with_id_filter(self, collection,
                                            filter_specification_factory):
        flt_spec = filter_specification_factory.create_equal_to('id', 0)
        collection.filter = flt_spec
        self.__check_url(resource_to_url(collection),
                         schema='http', path='/my-entities/', params='',
                         query='q=id:equal-to:0')

    def test_resource_to_url_with_resource_filter(self, resource_repo,
                                                  collection,
                                                filter_specification_factory):
        parent_coll = resource_repo.get_collection(IMyEntityParent)
        parent = parent_coll['0']
        parent_url = resource_to_url(parent)
        flt_spec = \
            filter_specification_factory.create_equal_to('parent', parent)
        collection.filter = flt_spec
        self.__check_url(resource_to_url(collection),
                         schema='http', path='/my-entities/', params='',
                         query='q=parent:equal-to:"%s"' % parent_url)

    def test_resource_to_url_with_order(self, collection,
                                        order_specification_factory):
        ord_spec = order_specification_factory.create_ascending('id')
        collection.order = ord_spec
        self.__check_url(resource_to_url(collection),
                         schema='http', path='/my-entities/', params='',
                         query='sort=id:asc')

    def test_resource_to_url_with_multiple_order(self,
                                                 collection,
                                                 order_specification_factory):
        ord_spec_id = order_specification_factory.create_ascending('id')
        ord_spec_text = order_specification_factory.create_descending('text')
        ord_spec = \
            order_specification_factory.create_conjunction(ord_spec_id,
                                                           ord_spec_text)
        collection.order = ord_spec
        self.__check_url(resource_to_url(collection),
                         schema='http', path='/my-entities/', params='',
                         query='sort=id:asc~text:desc')

    def test_resource_to_url_nested(self, member, resource_repo):
        child_root_coll = resource_repo.get_collection(type(member.children))
        srvc = child_root_coll.__parent__
        resource_repo.set_collection_parent(child_root_coll, None)
        try:
            coll = member.children
            coll_url = resource_to_url(coll)
            self.__check_url(coll_url, path='/my-entities/0/children/',
                             query='')
            mb = coll['0']
            mb_url = resource_to_url(mb)
            self.__check_url(mb_url, path='/my-entities/0/children/0/',
                             query='')
        finally:
            resource_repo.set_collection_parent(child_root_coll, srvc)

    def test_url_to_resource(self, collection):
        coll_from_url = url_to_resource(self.base_url)
        assert collection['0'] == coll_from_url['0']

    @pytest.mark.parametrize('url,error,msg',
                             [('http://0.0.0.0:6543/my-foos/', KeyError,
                               'has no subelement my-foos'),
                              ('http://0.0.0.0:6543/my-foos/', KeyError,
                               'has no subelement my-foos'),
                              ('http://0.0.0.0:6543/', ValueError,
                               'Traversal found non-resource object'),
                              (base_url + '?q=id|foo', ValueError,
                               'Expression parameters have errors'),
                              (base_url + '?sort=id|foo', ValueError,
                               'Expression parameters have errors'),
                              (base_url + '?start=0&size=a',
                               ValueError, 'must be a number.'),
                              (base_url + '?start=a&size=100',
                               ValueError, 'must be a number.'),
                              (base_url + '?start=-1&size=100',
                               ValueError,
                               'must be zero or a positive number.'),
                              (base_url + '?start=0&size=-100',
                               ValueError, 'must be a positive number.'),
                              ])
    def test_url_to_resource_invalid(self, url, error, msg):
        with pytest.raises(error) as cm:
            url_to_resource(url)
        assert str(cm.value).find(msg) != -1

    def test_url_to_resource_with_slice(self):
        coll_from_url = url_to_resource(self.base_url + '?size=1&start=0')
        # The length is not affected by the slice...
        assert len(coll_from_url) == 2
        # ... the actual number of members in the collection is.
        assert len(list(coll_from_url)) == 1

    def test_url_to_resource_with_filter(self):
        def _test(criterion, attr, value):
            coll_from_url = \
                        url_to_resource(self.base_url + '?q=%s' % criterion)
            mbs = list(coll_from_url)
            assert len(mbs) == 1
            assert getattr(mbs[0], attr) == value
        _test('id:equal-to:0', 'id', 0)
        _test('id:not-equal-to:0', 'id', 1)
        _test('text:starts-with:"foo"', 'text', 'foo0')
        _test('text:ends-with:"o1"', 'text', 'too1')
        _test('text:contains:"o0"', 'text', 'foo0')
        _test('text:not-contains:"o0"', 'text', 'too1')
        _test('text:contained:"foo0"', 'text', 'foo0')
        _test('text:not-contained:"foo0"', 'text', 'too1')
        _test('id:less-than:1', 'id', 0)
        _test('id:less-than-or-equal-to:0', 'id', 0)
        _test('id:greater-than:0', 'id', 1)
        _test('id:greater-than-or-equal-to:1', 'id', 1)
        _test('id:in-range:0-0', 'id', 0)

    def test_url_to_resource_with_filter_no_values_raises_error(self):
        pytest.raises(ValueError,
                           url_to_resource, self.base_url + '?q=id:equal-to:')

    def test_url_to_resource_with_complex_filter(self):
        criterion = '(id:equal-to:0 and text:equal-to:"foo0") or ' \
                    '(id:equal-to:1 and text:equal-to:"too1")'
        coll_from_url = \
                    url_to_resource(self.base_url + '?q=%s' % criterion)
        mbs = list(coll_from_url)
        assert len(mbs) == 2

    def test_url_to_resource_with_order(self):
        coll_from_url = url_to_resource(self.base_url + '?sort=id:asc')
        assert len(coll_from_url) == 2
        assert list(coll_from_url)[-1].id == 1

    def test_url_to_resource_with_multiple_order(self):
        coll_from_url = url_to_resource(self.base_url +
                                        '?sort=id:asc~text:desc')
        assert len(coll_from_url) == 2
        assert list(coll_from_url)[-1].id == 1

    def test_url_to_resource_with_multiple_filter(self):
        criteria = 'id:less-than:1~id:less-than-or-equal-to:1'
        coll_from_url = url_to_resource(self.base_url + '?q=%s' % criteria)
        assert len(coll_from_url) == 1

    def test_url_to_resource_with_multiple_criteria_one_empty(self):
        criteria = 'id:less-than:1~'
        coll_from_url = url_to_resource(self.base_url + '?q=%s' % criteria)
        assert len(coll_from_url) == 1

    def test_url_to_resource_with_multiple_values(self):
        criteria = 'id:equal-to:0,1'
        coll_from_url = url_to_resource(self.base_url + '?q=%s' % criteria)
        assert len(coll_from_url) == 2

    def test_url_to_resource_with_multiple_values_one_empty(self):
        criteria = 'id:equal-to:0,'
        coll_from_url = url_to_resource(self.base_url + '?q=%s' % criteria)
        assert len(coll_from_url) == 1

    def test_url_to_resource_with_multiple_string_values_one_empty(self):
        criteria = 'text:starts-with:"foo",""'
        coll_from_url = url_to_resource(self.base_url + '?q=%s' % criteria)
        assert len(coll_from_url) == 1

    def test_url_to_resource_with_link(self):
        criterion = 'parent:equal-to:"%s/my-entity-parents/0/"' % self.app_url
        coll_from_url = url_to_resource(self.base_url + '?q=%s' % criterion)
        assert len(coll_from_url) == 1

#    def test_url_to_resource_with_link_and_other(self):
#        criterion1 = 'parent:equal-to:%s/my-entity-parents/0/' % self.app_url
#        criterion2 = 'id:equal-to:0'
#        coll_from_url = url_to_resource(self.base_url +
#                                        '?q=%s~%s' % (criterion1, criterion2))
#        assertlen(coll_from_url), 1)

    def test_url_to_resource_with_link_and_other(self):
        criterion1 = 'parent:equal-to:"%s/my-entity-parents/0/"' \
                     % self.app_url
        criterion2 = 'id:equal-to:0'
        coll_from_url = url_to_resource(self.base_url +
                                        '?q=%s~%s' % (criterion1, criterion2))
        assert len(coll_from_url) == 1

    def test_two_urls(self):
        par_url = self.app_url + '/my-entity-parents/'
        criteria = 'parent:equal-to:"%s","%s"' \
                   % (par_url + '0/', par_url + '1/')
        url = self.base_url + '?q=%s' % criteria
        coll_from_url = url_to_resource(url)
        assert len(coll_from_url) == 2

    def test_url_to_resource_contained_with_simple_collection_link(self):
        nested_url = self.app_url \
                     + '/my-entity-parents/?q=id:less-than:1'
        url = self.app_url + '/my-entities/?q=parent:contained:' \
              + '"' + nested_url + '"'
        coll_from_url = url_to_resource(url)
        assert len(coll_from_url) == 1

    def test_url_to_resource_contained_with_complex_collection_link(self):
        for op, crit, num in zip((' and ', '~'),
                                 ('id:greater-than:0',
                                  'text:not-equal-to:"foo0"'),
                                 (1, 2)):
            nested_url = self.app_url \
                         + '/my-entity-parents/?q=id:less-than:2' \
                         + op \
                         + crit
            url = self.app_url + '/my-entities/?q=parent:contained:' \
                  + "'" + nested_url + "'"
            coll_from_url = url_to_resource(url)
            assert len(coll_from_url) == num

    def test_url_to_resource_contained_with_grouped_collection_link(self):
        url = self.app_url + '/my-entities/' \
              + '?q=(parent:contained:"' \
              + self.app_url \
              + '/my-entity-parents/?q=id:less-than:3") ' \
              + 'and text:not-equal-to:"foo0"'
        coll_from_url = url_to_resource(url)
        assert len(coll_from_url) == 1

    def test_nested_member_url_with_query_string_fail(self):
        par_url = self.app_url + '/my-entity-parents/1/'
        criteria = 'parent:equal-to:%s?q=id:equal-to:0' % par_url
        url = self.base_url + '?q=%s' % criteria
        pytest.raises(ValueError, url_to_resource, url)

    def test_url_to_resource_invalid_traversal_object(self, monkeypatch):
        monkeypatch.setattr(Service, '__getitem__',
                            classmethod(lambda cls, item: 1))
        url = self.app_url + '/foo'
        with pytest.raises(ValueError) as cm:
            url_to_resource(url)
        exc_msg = 'Traversal found non-resource object'
        assert str(cm.value).startswith(exc_msg)

    def __check_url(self, url,
                    schema=None, path=None, params=None, query=None):
        urlp = urlparse.urlparse(url)
        if not schema is None:
            assert urlp.scheme == schema # pylint: disable=E1101
        if not path is None:
            assert urlp.path == path # pylint: disable=E1101
        if not params is None:
            assert urlp.params == params # pylint: disable=E1101
        if not query is None:
            # We can not rely on the order of query parameters returned by
            # urlparse, so we compare the sets of parameters.
            assert set(urlp.query.split('&')) == \
                       set(query.split('&')) # pylint: disable=E1101


class TestUrlNoRdb(_TestUrl):
    config_file_name = 'configure_no_rdb.zcml'


@pytest.mark.usefixtures("rdb")
class TestUrlRdb(_TestUrl):
    config_file_name = 'configure.zcml'



"""
Resource relationship class.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 12, 2013.
"""
from everest.relationship import RELATIONSHIP_DIRECTIONS
from everest.relationship import Relationship

__docformat__ = 'reStructuredText en'
__all__ = ['ResourceRelationship',
           ]


class ResourceRelationship(Relationship):
    """
    Relationship between resources.
    """
    def __init__(self, relator, descriptor,
                 direction=RELATIONSHIP_DIRECTIONS.BIDIRECTIONAL):
        Relationship.__init__(self, relator, descriptor, direction)
        self.__domain_relationship = None

    @property
    def domain_relationship(self):
        """
        Returns a domain relationship equivalent with this resource
        relationship.
        """
        if self.__domain_relationship is None:
            ent = self.relator.get_entity()
            self.__domain_relationship = \
                    self.descriptor.make_relationship(ent)
        return self.__domain_relationship

    def add(self, related, direction=None, safe=False):
        self.domain_relationship.add(related.get_entity(),
                                     direction=direction,
                                     safe=safe)

    def remove(self, related, direction=None, safe=False):
        self.domain_relationship.remove(related.get_entity(),
                                        direction=direction,
                                        safe=safe)

    def _get_specification_attributes(self):
        return self.descriptor.resource_attr, self.descriptor.resource_backref





"""
Resource link.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 29, 2011.
"""
from everest.resources.interfaces import IResourceLink
from everest.resources.utils import resource_to_url
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['Link',
           ]


@implementer(IResourceLink)
class Link(object):
    """
    A resource link.

    :note: The URL for the linked resource is created lazily; at
      instantiation time, we may not have a request to generate the URL.
    """
    def __init__(self, linked_resource, rel,
                 type=None, title=None, length=None): # pylint: disable=W0622
        self.__linked_resource = linked_resource
        self.rel = rel
        self.type = type
        self.title = title
        self.length = length

    @property
    def href(self):
        return resource_to_url(self.__linked_resource)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 18, 2013.
"""
import codecs
from csv import DictReader
from csv import reader

from pyramid.compat import PY3
from pyramid.compat import binary_type


__docformat__ = 'reStructuredText en'
__all__ = ['csv_reader',
           'izip',
           'open_text',
           'parse_qsl',
           'BytesIO',
           'CsvDictReader',
           ]


# pylint: disable=E0611
if PY3:
    izip = zip
else:
    from itertools import izip

if PY3:
    from urllib import parse
    parse_qsl = parse.parse_qsl
else:
    import urlparse

    parse_qsl = urlparse.parse_qsl


if PY3:
    from io import BytesIO
else:
    from StringIO import StringIO as BytesIO


if PY3:
    def open_text(filename):
        return open(filename, 'w', newline='')
else:
    def open_text(filename):
        return open(filename, 'wb')


if PY3:
    csv_reader = reader
    CsvDictReader = DictReader
else:
    from csv import excel
    from pyramid.compat import text_type

    # The following classes are used to support encodings in CSV files under
    # Python 2.7.x . Adapted from the csv module documentation.

    class Utf8Recoder:
        def __init__(self, f, encoding):
            self.reader = codecs.getreader(encoding)(f)

        def __iter__(self):
            return self

        def next(self):
            return self.reader.next().encode('utf-8')

        def read(self, size):
            return self.reader.read(size).encode('utf-8')


    class UnicodeReader(object):
        def __init__(self, f, dialect=excel, encoding='utf-8', **kwargs):
            if not encoding is None:
                f = Utf8Recoder(f, encoding)
            self.reader = reader(f, dialect=dialect, **kwargs)

        def next(self):
            row = next(self.reader)
            return [text_type(cell, 'utf-8')
                    if isinstance(cell, binary_type) else cell
                    for cell in row]

        __next__ = next

        @property
        def line_num(self):
            return self.reader.line_num

        def __iter__(self):
            return self


    class UnicodeDictReader(DictReader):
        def __init__(self, f, fieldnames=None, restkey=None, restval=None,
                     dialect="excel", encoding='utf-8', *args, **kw):
            DictReader.__init__(self, f, fieldnames=fieldnames,
                                restkey=restkey, restval=restval,
                                dialect=dialect, *args, **kw)
            if not encoding is None:
                f = Utf8Recoder(f, encoding=encoding)
            # Replace the reader with our unicode-enabled reader.
            self.reader = reader(f, dialect=dialect, *args, **kw)


    csv_reader = reader # UnicodeReader
    CsvDictReader = DictReader # UnicodeDictReader

# pylint: enable=E0611



"""
Resource data tree traversal.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 25, 2012.
"""
from collections import OrderedDict
from everest.attributes import is_terminal_attribute
from everest.constants import RELATIONSHIP_DIRECTIONS
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.constants import RESOURCE_KINDS
from everest.entities.utils import get_entity_class
from everest.interfaces import IDataTraversalProxyFactory
from everest.representers.attributes import MappedAttributeKey
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import WRITE_AS_LINK_OPTION
from everest.representers.config import WRITE_MEMBERS_AS_LINK_OPTION
from everest.representers.interfaces import ICollectionDataElement
from everest.representers.interfaces import ILinkedDataElement
from everest.representers.interfaces import IMemberDataElement
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from everest.traversal import ConvertingDataTraversalProxyMixin
from everest.traversal import DataTraversalProxy
from everest.traversal import DataTraversalProxyAdapter
from everest.utils import set_nested_attribute
from pyramid.compat import iteritems_
from pyramid.threadlocal import get_current_registry
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['DataElementBuilderRepresentationDataVisitor',
           'DataElementBuilderResourceDataVisitorBase',
           'DataElementBuilderResourceTreeVisitor',
           'DataElementDataTraversalProxy',
           'DataElementDataTraversalProxyAdapter',
           'DataElementTreeTraverser',
           'DataTreeTraverser',
           'DataElementTreeTraverser',
           'ResourceDataTreeTraverser',
           'ResourceDataVisitor',
           'ResourceTreeTraverser',
           ]


class ResourceDataVisitor(object):
    def visit_member(self, attribute_key, attribute, member_node, member_data,
                     is_link_node, parent_data, index=None):
        """
        Visits a member node in a resource data tree.

        :param tuple attribute_key: tuple containing the attribute tokens
          identifying the member node's position in the resource data tree.
        :param attribute: mapped attribute holding information about the
          member node's name (in the parent) and type etc.
        :type attribute:
          :class:`everest.representers.attributes.MappedAttribute`
        :param member_node: the node holding resource data. This is either a
          resource instance (when using a :class:`ResourceTreeTraverser` on
          a tree of resources) or a data element instance (when using a
          :class:`DataElementTreeTraverser` on a data element tree.
        :param dict member_data: dictionary holding all member data
          extracted during traversal (with mapped attributes as keys).
        :param bool is_link_node: indicates if the given member node is a link.
        :param dict parent_data: dictionary holding all parent data extracted
          during traversal (with mapped attributes as keys).
        :param int index: this indicates a member node's index in a collection
          parent node. If the parent node is a member node, it will be `None`.
        """
        raise NotImplementedError('Abstract method.')

    def visit_collection(self, attribute_key, attribute, collection_node,
                         collection_data, is_link_node, parent_data):
        """
        Visits a collection node in a resource data tree.

        :param tuple attribute_key: tuple containing the attribute tokens
          identifying the collection node's position in the resource data
          tree.
        :param attribute: mapped attribute holding information about the
          collection node's name (in the parent) and type etc.
        :type attribute:
          :class:`everest.representers.attributes.MappedAttribute`
        :param collection_node: the node holding resource data. This is either
          a resource instance (when using a :class:`ResourceTreeTraverser` on
          a tree of resources) or a data element instance (when using a
          :class:`DataElementTreeTraverser` on a data element tree.
        :param dict collection_data: dictionary mapping member index to member
          data for each member in the visited collection.
        :param bool is_link_node: indicates if the given member node is a link.
        :param dict parent_data: dictionary holding all parent data extracted
          during traversal (with mapped attributes as keys).
        """
        raise NotImplementedError('Abstract method.')


class DataElementBuilderResourceDataVisitorBase(ResourceDataVisitor):
    """
    Abstract base class for visitors creating a data element from resource
    data.
    """
    def __init__(self, mapping):
        ResourceDataVisitor.__init__(self)
        self._mapping = mapping
        self.__data_el = None

    def visit_member(self, attribute_key, attribute, member_node, member_data,
                     is_link_node, parent_data, index=None):
        if is_link_node:
            mb_data_el = self._create_link_data_element(attribute,
                                                        member_node)
        else:
            mb_data_el = self._create_member_data_element(attribute,
                                                          member_node)
            # Process attributes.
            for attr, value in iteritems_(member_data):
                if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                    self._set_terminal_attribute(mb_data_el, attr, value)
                else:
                    mb_data_el.set_nested(attr, value)
        if not index is None:
            # Collection member. Store in parent data with index as key.
            parent_data[index] = mb_data_el
        elif len(attribute_key) == 0:
            # Top level - store root data element.
            self.__data_el = mb_data_el
        else:
            # Nested member. Store in parent data with attribute as key.
            parent_data[attribute] = mb_data_el

    def visit_collection(self, attribute_key, attribute, collection_node,
                         collection_data, is_link_node, parent_data):
        if is_link_node:
            coll_data_el = self._create_link_data_element(attribute,
                                                          collection_node)
        else:
            coll_data_el = \
                self._create_collection_data_element(attribute,
                                                     collection_node)
            for item in sorted(collection_data.items()):
                coll_data_el.add_member(item[1])
        if len(attribute_key) == 0: # Top level.
            self.__data_el = coll_data_el
        else:
            parent_data[attribute] = coll_data_el

    @property
    def data_element(self):
        return self.__data_el

    def _create_member_data_element(self, attribute, member_node):
        raise NotImplementedError('Abstract method.')

    def _create_collection_data_element(self, attribute, collection_node):
        raise NotImplementedError('Abstract method.')

    def _create_link_data_element(self, attribute, member_node):
        raise NotImplementedError('Abstract method.')

    def _set_terminal_attribute(self, data_element, attribute, value):
        raise NotImplementedError('Abstract method.')


class DataElementBuilderRepresentationDataVisitor(
                                    DataElementBuilderResourceDataVisitorBase):
    def _create_member_data_element(self, attribute, member_node):
        if not attribute is None:
            mb_cls = get_member_class(attribute.value_type)
        else:
            mb_cls = get_member_class(self._mapping.mapped_class)
        return self._mapping.create_data_element(mapped_class=mb_cls)

    def _create_collection_data_element(self, attribute, collection_node):
        if not attribute is None:
            coll_cls = get_collection_class(attribute.value_type)
        else:
            coll_cls = get_collection_class(self._mapping.mapped_class)
        return self._mapping.create_data_element(mapped_class=coll_cls)

    def _create_link_data_element(self, attribute, member_node):
        if attribute.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
            kind = RESOURCE_KINDS.MEMBER
            rc_cls = get_member_class(attribute.value_type)
        else:
            kind = RESOURCE_KINDS.COLLECTION
            rc_cls = get_collection_class(attribute.value_type)
        return self._mapping.create_linked_data_element(
                                                member_node, kind,
                                                relation=rc_cls.relation,
                                                title=rc_cls.title)

    def _set_terminal_attribute(self, data_element, attribute, value):
        data_element.set_terminal_converted(attribute, value)


class DataElementBuilderResourceTreeVisitor(
                                    DataElementBuilderResourceDataVisitorBase):
    def _create_member_data_element(self, attribute, member_node):
        return self._mapping.create_data_element_from_resource(member_node)

    def _create_collection_data_element(self, attribute, collection_node):
        return \
            self._mapping.create_data_element_from_resource(collection_node)

    def _create_link_data_element(self, attribute, member_node):
        return \
          self._mapping.create_linked_data_element_from_resource(member_node)

    def _set_terminal_attribute(self, data_element, attribute, value):
        data_element.set_terminal(attribute, value)


class DataTreeTraverser(object):
    """
    Abstract base class for data tree traversers.
    """
    def __init__(self, root):
        self.__root = root

    def run(self, visitor):
        """
        Runs this traverser.
        """
        self._dispatch(MappedAttributeKey(()), None, self.__root, None,
                       visitor)

    def _traverse_collection(self, attr_key, attr, collection_node,
                             parent_data, visitor):
        collection_data = {}
        is_link_node = \
            self._is_link_node(collection_node, attr) \
            and not (not attr is None and
                     attr.options.get(WRITE_MEMBERS_AS_LINK_OPTION) is True)
        if not is_link_node:
            all_mb_nodes = self._get_node_members(collection_node)
            for idx, mb_node in enumerate(all_mb_nodes):
                self._traverse_member(attr_key, attr, mb_node, collection_data,
                                      visitor, index=idx)
        visitor.visit_collection(attr_key, attr, collection_node,
                                 collection_data, is_link_node, parent_data)

    def _traverse_member(self, attr_key, attr, member_node, parent_data,
                         visitor, index=None):
        raise NotImplementedError('Abstract method.')

    def _is_link_node(self, node, attr):
        raise NotImplementedError('Abstract method.')

    def _get_node_members(self, node):
        raise NotImplementedError('Abstract method.')

    def _dispatch(self, attr_key, attr, node, parent_data, visitor):
        raise NotImplementedError('Abstract method.')


class ResourceDataTreeTraverser(DataTreeTraverser):
    """
    Abstract base class for resource data tree traversers.
    """
    def __init__(self, root, mapping, ignore_none_values=True):
        DataTreeTraverser.__init__(self, root)
        self._mapping = mapping
        self.__ignore_none_values = ignore_none_values

    def _traverse_member(self, attr_key, attr, member_node, parent_data,
                         visitor, index=None):
        member_data = OrderedDict()
        is_link_node = \
            self._is_link_node(member_node, attr) \
            or (not index is None and not attr is None and
                attr.options.get(WRITE_MEMBERS_AS_LINK_OPTION) is True)
        # Ignore links for traversal.
        if not is_link_node:
            if not attr is None:
                node_type = get_member_class(attr.value_type)
            else:
                node_type = self._get_node_type(member_node)
            for mb_attr in self._mapping.attribute_iterator(node_type,
                                                            attr_key):
                ignore_opt = self._get_ignore_option(mb_attr)
                if mb_attr.should_ignore(attr_key):
                    continue
                if mb_attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                    # Terminal attribute - extract.
                    value = self._get_node_terminal(member_node, mb_attr)
                    if value is None and self.__ignore_none_values:
                        # We ignore None attribute values when reading
                        # representations.
                        continue
                    member_data[mb_attr] = value
                else:
                    # Nested attribute - traverse.
                    nested_node = self._get_node_nested(member_node,
                                                        mb_attr)
                    if nested_node is None:
                        # Stop condition - the given data element does not
                        # contain a nested attribute of the given mapped
                        # name.
                        continue
                    nested_attr_key = attr_key + (mb_attr,)
                    if ignore_opt is False:
                        # The offset in the attribute key ensures that
                        # the defaults for ignoring attributes of the
                        # nested attribute can be retrieved correctly.
                        nested_attr_key.offset = len(nested_attr_key)
                    self._dispatch(nested_attr_key, mb_attr, nested_node,
                                   member_data, visitor)
        visitor.visit_member(attr_key, attr, member_node, member_data,
                             is_link_node, parent_data, index=index)

    def _get_node_type(self, node):
        raise NotImplementedError('Abstract method.')

    def _get_node_terminal(self, node, attr):
        raise NotImplementedError('Abstract method.')

    def _get_node_nested(self, node, attr):
        raise NotImplementedError('Abstract method.')

    def _get_ignore_option(self, attr):
        return attr.options.get(IGNORE_OPTION)


class DataElementTreeTraverser(ResourceDataTreeTraverser):
    """
    Traverser for data element trees.

    This legacy traverser is only needed to generate a representation from
    a data element tree.
    """
    def __init__(self, root, mapping,
                 ignore_none_values=True):
        ResourceDataTreeTraverser.__init__(
                                    self, root, mapping,
                                    ignore_none_values=ignore_none_values)

    def _dispatch(self, attr_key, attr, node, parent_data, visitor):
        ifcs = provided_by(node)
        if IMemberDataElement in ifcs:
            traverse_fn = self._traverse_member
        elif ICollectionDataElement in ifcs:
            traverse_fn = self._traverse_collection
        elif ILinkedDataElement in ifcs:
            kind = node.get_kind()
            if kind == RESOURCE_KINDS.MEMBER:
                traverse_fn = self._traverse_member
            else: # kind == RESOURCE_KINDS.COLLECTION
                traverse_fn = self._traverse_collection
        else:
            raise ValueError('Need MEMBER or COLLECTION data element; found '
                             '"%s".' % node)
        traverse_fn(attr_key, attr, node, parent_data, visitor)

    def _is_link_node(self, node, attr): # pylint: disable=W0613
        return ILinkedDataElement in provided_by(node)

    def _get_node_members(self, node):
        return node.get_members()

    def _get_node_type(self, node):
        return node.mapping.mapped_class

    def _get_node_terminal(self, node, attr):
        return node.get_terminal_converted(attr)

    def _get_node_nested(self, node, attr):
        return node.get_nested(attr)


class ResourceTreeTraverser(ResourceDataTreeTraverser):
    """
    Mapping traverser for resource trees.
    """
    def __init__(self, root, mapping, ignore_none_values=True):
        ResourceDataTreeTraverser.__init__(
                                    self, root, mapping,
                                    ignore_none_values=ignore_none_values)

    def _dispatch(self, attr_key, attr, node, parent_data, visitor):
        ifcs = provided_by(node)
        if IMemberResource in ifcs:
            self._traverse_member(attr_key, attr, node, parent_data, visitor)
        elif ICollectionResource in ifcs:
            self._traverse_collection(attr_key, attr, node, parent_data,
                                      visitor)
        else:
            raise ValueError('Can only traverse objects that provide'
                             'IMemberResource or ICollectionResource '
                             '(key: %s).' % str(attr_key))

    def _get_node_type(self, node):
        return type(node)

    def _get_node_terminal(self, node, attr):
        return getattr(node, attr.name)

    def _get_node_nested(self, node, attr):
        return getattr(node, attr.name)

    def _get_node_members(self, node):
        return iter(node)

    def _is_link_node(self, node, attr):
        return not attr is None and \
               not attr.options.get(WRITE_AS_LINK_OPTION) is False


class DataElementDataTraversalProxy(ConvertingDataTraversalProxyMixin,
                                    DataTraversalProxy):
    def __init__(self, data, accessor, relationship_direction,
                 attribute_key=None, mapping=None):
        """
        :param mapping: resource attribute mapping. This needs to be passed
          along to all other, nested proxies generated by this one in the
          traversal process.
        :type mapping: :class:`everest.representers.mapping.Mapping`
        """
        DataTraversalProxy.__init__(self, data, accessor,
                                    relationship_direction)
        if attribute_key is None:
            attribute_key = MappedAttributeKey(())
        self.__attribute_key = attribute_key
        if mapping is None:
            mapping = data.mapping
        self.__mapping = mapping

    def get_id(self):
        id_attr = self.__mapping.get_attribute_map()['id']
        return self._data.get_terminal(id_attr)

    def _get_entity_type(self):
        return get_entity_class(self._data.mapping.mapped_class)

    def _get_proxy_options(self, attribute):
        return dict(attribute_key=self.__attribute_key + (attribute,),
                    mapping=self.__mapping)

    def _attribute_iterator(self):
        return self.__mapping.attribute_iterator(
                                mapped_class=self._data.mapping.mapped_class,
                                key=self.__attribute_key)

    def _get_relation_attribute_value(self, attribute):
        # We use the access by representation name here which triggers an
        # AttributeError if the attribute was not set.
        return self._data.get_attribute(attribute.repr_name)

    def _get_proxied_attribute_value(self, attribute):
        if not is_terminal_attribute(attribute):
            val = self._get_relatee(attribute)
        else:
            val = self._data.get_attribute(attribute.repr_name)
        return val

    def _make_accessor(self, value_type):
        raise NotImplementedError('Not available for data element proxies.')

    def _convert_to_entity(self):
        init_map = {}
        nested_map = {}
        mapped_class = self._data.mapping.mapped_class
        for attr in \
            self.__mapping.attribute_iterator(mapped_class=mapped_class,
                                              key=self.__attribute_key):
            try:
                val = self._get_proxied_attribute_value(attr)
            except AttributeError:
                continue
            else:
                attr_name = attr.entity_attr
                if not '.' in attr_name:
                    init_map[attr_name] = val
                else:
                    nested_map[attr_name] = val
        ent_cls = get_entity_class(mapped_class)
        entity = ent_cls.create_from_data(init_map)
        for nested_name, nested_value in iteritems_(nested_map):
            set_nested_attribute(entity, nested_name, nested_value)
        return entity

    def __str__(self):
        return "%s(id=%s)" % (self._data.__class__.__name__,
                              self.get_id())


class DataElementDataTraversalProxyAdapter(DataTraversalProxyAdapter):
    proxy_class = DataElementDataTraversalProxy

    def make_source_proxy(self, options=None):
        rel_drct = RELATIONSHIP_DIRECTIONS.REVERSE
        if options is None:
            options = {}
        options['mapping'] = self._data.mapping
        return self.make_proxy(None, rel_drct, options)

    def make_target_proxy(self, accessor,
                          manage_back_references=True, options=None):
        rel_drct = RELATIONSHIP_DIRECTIONS.BIDIRECTIONAL
        if not manage_back_references:
            rel_drct &= ~RELATIONSHIP_DIRECTIONS.REVERSE
        return self.make_proxy(accessor, rel_drct, options=options)

    def make_proxy(self, accessor, relationship_direction, options=None):
        if ICollectionDataElement.providedBy(self._data): # pylint:disable=E1101
            reg = get_current_registry()
            prx_fac = reg.getUtility(IDataTraversalProxyFactory)
            prx = prx_fac.make_proxy(list(self._data.get_members()),
                                     accessor, relationship_direction,
                                     options=options)
        else:
            prx = DataTraversalProxyAdapter.make_proxy(self, accessor,
                                                       relationship_direction,
                                                       options=options)
        return prx



"""
Repository utilities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 17, 2013.
"""
from everest.repositories.interfaces import IRepository
from pyramid.threadlocal import get_current_registry
from threading import Lock
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface.interfaces import IInterface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['GlobalObjectManager',
           'commit_veto',
           'get_engine',
           'is_engine_initialized',
           'reset_engines',
           'set_engine',
           ]


class GlobalObjectManager(object):
    _globs = None
    _lock = None

    @classmethod
    def set(cls, key, obj):
        """
        Sets the given object as global object for the given key.
        """
        with cls._lock:
            if not cls._globs.get(key) is None:
                raise ValueError('Duplicate key "%s".' % key)
            cls._globs[key] = obj
        return cls._globs[key]

    @classmethod
    def get(cls, key):
        """
        Returns the global object for the given key.

        :raises KeyError: if no global object was initialized for the given
          key.
        """
        with cls._lock:
            return cls._globs[key]

    @classmethod
    def is_initialized(cls, key):
        """
        Checks if a global object with the given key has been initialized.
        """
        with cls._lock:
            return not cls._globs.get(key) is None

    @classmethod
    def reset(cls):
        """
        Discards all global objects held by this manager.
        """
        with cls._lock:
            cls._globs.clear()


class _DbEngineManager(GlobalObjectManager):
    _globs = {}
    _lock = Lock()

get_engine = _DbEngineManager.get
set_engine = _DbEngineManager.set
is_engine_initialized = _DbEngineManager.is_initialized
reset_engines = _DbEngineManager.reset


def as_repository(resource):
    """
    Adapts the given registered resource to its configured repository.

    :return: object implementing
      :class:`everest.repositories.interfaces.IRepository`.
    """
    reg = get_current_registry()
    if IInterface in provided_by(resource):
        resource = reg.getUtility(resource, name='collection-class')
    return reg.getAdapter(resource, IRepository)


def commit_veto(request, response): # unused request arg pylint: disable=W0613
    """
    Strict commit veto to use with the transaction manager.

    Unlike the default commit veto supplied with the transaction manager,
    this will veto all commits for HTTP status codes other than 2xx unless
    a commit is explicitly requested by setting the "x-tm" response header to
    "commit". As with the default commit veto, the commit is always vetoed if
    the "x-tm" response header is set to anything other than "commit".
    """
    tm_header = response.headers.get('x-tm')
    if not tm_header is None:
        result = tm_header != 'commit'
    else:
        result = not response.status.startswith('2') \
                 and not tm_header == 'commit'
    return result



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 19, 2013.
"""
from everest.testing import EntityTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.entities import MyEntityChild
from everest.repositories.state import EntityState
from everest.resources.descriptors import terminal_attribute
from mock import MagicMock

__docformat__ = 'reStructuredText en'
__all__ = ['EntityStateTestCase',
           ]


class EntityStateTestCase(EntityTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_state_data(self):
        data = dict(text='FOO',
                    number=-1,
                    parent=MyEntityParent(),
                    children=[MyEntityChild()])
        entity = MyEntity(**data)
        # We don't want to test the required unit of work here.
        uow = MagicMock()
        self.assert_raises(ValueError, EntityState.get_state, entity)
        entity.__everest__ = EntityState(entity, uow)
        state_data = EntityState.get_state(entity).data
        for attr, value in state_data.items():
            if attr.resource_attr == 'number':
                number_attr = attr
            elif attr.resource_attr == 'parent':
                parent_attr = attr
            elif attr.resource_attr == 'parent_text':
                parent_text_attr = attr
            if attr.resource_attr in data:
                self.assert_equal(value, data[attr.resource_attr])
        new_number = -2
        state_data[number_attr] = new_number
        EntityState.get_state(entity).data = state_data
        self.assert_equal(entity.number, new_number)
        new_entity = MyEntity()
        self.assert_not_equal(new_entity.number, new_number)
        new_entity.__everest__ = EntityState(new_entity, uow)
        EntityState.transfer_state_data(entity, new_entity)
        self.assert_equal(new_entity.number, new_number)
        # Make setting invalid attribute fail.
        invalid_number_attr = terminal_attribute(str, 'grmbl')
        del state_data[number_attr]
        state_data[invalid_number_attr] = -2
        with self.assert_raises(ValueError) as cm:
            EntityState.get_state(entity).data = state_data
        self.assert_true(cm.exception.args[0].startswith('Can not set'))
        # Make set nested attribute fail.
        entity.parent = None
        del state_data[invalid_number_attr]
        del state_data[parent_attr]
        state_data[parent_text_attr] = 'FOO PARENT'
        state = EntityState.get_state(entity)
        self.assert_raises(AttributeError, setattr, state, 'data', state_data)





"""
Entity attributes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 2, 2011.
"""
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.interfaces import IEntity
from everest.entities.utils import get_entity_class
from functools import wraps
from pyramid.compat import itervalues_
from zope.interface import implementedBy as implemented_by # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['get_domain_class_attribute_iterator',
           'get_domain_class_attribute',
           'get_domain_class_attribute_iterator',
           'get_domain_class_attribute_names',
           'get_domain_class_attributes',
           'get_domain_class_relationship_attribute_iterator',
           'get_domain_class_member_attribute_iterator',
           'get_domain_class_terminal_attribute_iterator',
           'is_domain_class_collection_attribute',
           'is_domain_class_domain_attribute',
           'is_domain_class_member_attribute',
           'is_domain_class_terminal_attribute',
           ]


def _arg_to_entity_class(func):
    @wraps(func)
    def wrap(ent, *args):
        if isinstance(ent, type) and IEntity in implemented_by(ent):
            ent_cls = ent
        else:
            ent_cls = get_entity_class(ent)
        return func(ent_cls, *args)
    return wrap


def is_domain_class_terminal_attribute(ent, attr_name):
    """
    Checks if the given attribute name is a terminal attribute of the given
    registered resource.
    """
    attr = get_domain_class_attribute(ent, attr_name)
    return attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL


def is_domain_class_member_attribute(ent, attr_name):
    """
    Checks if the given attribute name is a entity attribute of the given
    registered resource.
    """
    attr = get_domain_class_attribute(ent, attr_name)
    return attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER


def is_domain_class_collection_attribute(ent, attr_name):
    """
    Checks if the given attribute name is a aggregate attribute of the given
    registered resource.
    """
    attr = get_domain_class_attribute(ent, attr_name)
    return attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION


def is_domain_class_domain_attribute(ent, attr_name):
    """
    Checks if the given attribute name is a resource attribute (i.e., either
    a member or a aggregate attribute) of the given registered resource.
    """
    attr = get_domain_class_attribute(ent, attr_name)
    return attr != RESOURCE_ATTRIBUTE_KINDS.TERMINAL


@_arg_to_entity_class
def get_domain_class_attribute_names(ent):
    """
    Returns all attribute names of the given registered resource.
    """
    return ent.__everest_attributes__.keys()


@_arg_to_entity_class
def get_domain_class_attributes(ent):
    """
    Returns a dictionary mapping the attribute names of the given
    registered resource to :class:`ResourceAttribute` instances.
    """
    return ent.__everest_attributes__


@_arg_to_entity_class
def get_domain_class_attribute(ent, name):
    """
    Returns the specified attribute from the map of all collected attributes
    for the given registered resource or `None`, if the attribute could not
    be found.
    """
    return ent.__everest_attributes__.get(name)


@_arg_to_entity_class
def get_domain_class_attribute_iterator(ent):
    """
    Returns an iterator over all attributes in the given registered
    resource.
    """
    return itervalues_(ent.__everest_attributes__)


@_arg_to_entity_class
def get_domain_class_terminal_attribute_iterator(ent):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(ent.__everest_attributes__):
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
            yield attr


@_arg_to_entity_class
def get_domain_class_relationship_attribute_iterator(ent):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(ent.__everest_attributes__):
        if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
            yield attr


@_arg_to_entity_class
def get_domain_class_member_attribute_iterator(ent):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(ent.__everest_attributes__):
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
            yield attr


@_arg_to_entity_class
def get_domain_class_collection_attribute_iterator(ent):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(ent.__everest_attributes__):
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
            yield attr



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 17, 2014.
"""
import pytest

from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.testing import create_entity
from everest.representers.utils import as_representer


__docformat__ = 'reStructuredText en'
__all__ = ['collection',
           ]


@pytest.fixture
def collection(resource_repo, entity_id1=0, entity_id2=1):
    my_entity1 = create_entity(entity_id=None, entity_text='foo0')
    my_entity2 = create_entity(entity_id=None, entity_text='too1')
    coll = resource_repo.get_collection(IMyEntity)
    coll.create_member(my_entity1)
    coll.create_member(my_entity2)
    my_entity1.id = entity_id1
    my_entity1.parent.id = entity_id1
    my_entity1.children[0].id = entity_id1
    my_entity1.children[0].children[0].id = entity_id1
    my_entity2.id = entity_id2
    my_entity2.parent.id = entity_id2
    my_entity2.children[0].id = entity_id2
    my_entity2.children[0].children[0].id = entity_id2
    return coll


@pytest.fixture
def member(collection): # pylint: disable=W0621
    return next(iter(collection))


@pytest.fixture
def representer(request, collection): # pylint: disable=W0621
    cnt_type = request.cls.content_type
    return as_representer(collection, cnt_type)


@pytest.fixture
def member_representer(request, member): # pylint: disable=W0621
    cnt_type = request.cls.content_type
    return as_representer(member, cnt_type)


@pytest.fixture
def mapping(representer): # pylint: disable=W0621
    return representer._mapping # accessing protected pylint: disable=W0212


@pytest.fixture
def member_mapping(member_representer): # pylint: disable=W0621
    return member_representer._mapping # accessing protected pylint: disable=W0212




"""
JSON representers.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Aug 29, 2012.
"""
from __future__ import absolute_import # Makes the import below absolute

from collections import OrderedDict
import datetime
from json import dumps
from json import loads

from pyramid.compat import binary_type
from pyramid.compat import bytes_
from pyramid.compat import iteritems_
from pyramid.compat import string_types

from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.mime import JsonMime
from everest.representers.base import MappingResourceRepresenter
from everest.representers.base import RepresentationGenerator
from everest.representers.base import RepresentationParser
from everest.representers.config import RepresenterConfiguration
from everest.representers.converters import BooleanConverter
from everest.representers.converters import ConverterRegistry
from everest.representers.converters import DateTimeConverter
from everest.representers.converters import NoOpConverter
from everest.representers.dataelements import SimpleCollectionDataElement
from everest.representers.dataelements import SimpleLinkedDataElement
from everest.representers.dataelements import SimpleMemberDataElement
from everest.representers.interfaces import IRepresentationConverter
from everest.representers.mapping import SimpleMappingRegistry
from everest.representers.traversal import DataElementTreeTraverser
from everest.representers.traversal import ResourceDataTreeTraverser
from everest.representers.traversal import ResourceDataVisitor
from everest.representers.traversal import \
                                DataElementBuilderRepresentationDataVisitor
from everest.resources.utils import get_member_class
from everest.resources.utils import get_resource_class_for_relation
from everest.resources.utils import is_resource_url
from zope.interface import provider # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['JsonCollectionDataElement',
           'JsonConverterRegistry',
           'JsonDataElementTreeVisitor',
           'JsonDataTreeTraverser',
           'JsonLinkedDataElement',
           'JsonMappingRegistry',
           'JsonMemberDataElement',
           'JsonRepresentationGenerator',
           'JsonRepresentationParser',
           'JsonRepresenterConfiguration',
           'JsonResourceRepresenter',
           ]


class JsonConverterRegistry(ConverterRegistry):
    pass


@provider(IRepresentationConverter)
class JsonBytesConverter(object):
    """
    The JSON decoder returns text which we have to encode for byte attributes.
    """

    @classmethod
    def from_representation(cls, value):
        return bytes_(value, encoding='utf-8')

    @classmethod
    def to_representation(cls, value):
        return value


JsonConverterRegistry.register(datetime.datetime, DateTimeConverter)
JsonConverterRegistry.register(bool, BooleanConverter)
JsonConverterRegistry.register(int, NoOpConverter)
JsonConverterRegistry.register(float, NoOpConverter)
JsonConverterRegistry.register(binary_type, JsonBytesConverter)


class JsonDataTreeTraverser(ResourceDataTreeTraverser):
    """
    Specialized traverser that extracts resource data from a tree of JSON data.
    """
    def _dispatch(self, attr_key, attr, node, parent_data, visitor):
        if isinstance(node, dict):
            traverse_fn = self._traverse_member
        elif isinstance(node, list):
            traverse_fn = self._traverse_collection
        else:
            if not isinstance(node, string_types):
                raise ValueError('Need dict (member), list (collection) '
                                 'or string (URL) type for JSON data, found '
                                 '"%s"' % type(node))
            if attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
                traverse_fn = self._traverse_member
            elif attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
                traverse_fn = self._traverse_collection
        traverse_fn(attr_key, attr, node, parent_data, visitor)

    def _get_node_type(self, node):
        relation = node.get('__jsonclass__')
        if not relation is None:
            tpe = get_resource_class_for_relation(relation)
        else:
            # In the absence of class hinting, the best we can do is to
            # look up the member class for the mapped class. For polymorphic
            # types, this will only work if a representer was initialized
            # for every derived class separately.
            tpe = get_member_class(self._mapping.mapped_class)
        return tpe

    def _get_node_terminal(self, node, attr):
        return node.get(attr.repr_name)

    def _get_node_nested(self, node, attr):
        nested_data = node.get(attr.repr_name)
        if isinstance(nested_data, dict): # Can also be None or a URL.
            json_class = nested_data.pop('__jsonclass__', None)
            if not json_class is None:
                rel = get_member_class(attr.value_type).relation
                if json_class != rel:
                    raise ValueError('Expected data for %s, got %s.'
                                     % (rel, json_class))
        return nested_data

    def _get_node_members(self, node):
        return node

    def _is_link_node(self, node, attr): # pylint: disable=W0613
        return isinstance(node, string_types) and is_resource_url(node)


class JsonRepresentationParser(RepresentationParser):
    """
    Implementation of a representation parser for JSON.
    """
    def run(self):
        json_data = loads(self._stream.read(),
                          encoding=self.get_option('encoding', 'utf-8'))
        trv = JsonDataTreeTraverser(json_data, self._mapping)
        vst = DataElementBuilderRepresentationDataVisitor(self._mapping)
        trv.run(vst)
        return vst.data_element


class JsonDataElementTreeVisitor(ResourceDataVisitor):
    """
    Visitor creating JSON representations from data element nodes.
    """
    def __init__(self):
        ResourceDataVisitor.__init__(self)
        self.__json_data = None

    def visit_member(self, attribute_key, attribute, member_node, member_data,
                     is_link_node, parent_data, index=None):
        if is_link_node:
            mb_data = member_node.get_url()
        else:
            # Using an ordered dict gives us reproducible representations.
            mb_data = OrderedDict()
            for attr, value in iteritems_(member_data):
#                if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                mb_data[attr.repr_name] = value
            # Use the relation for class hinting.
            mb_cls = member_node.mapping.mapped_class
            mb_data['__jsonclass__'] = mb_cls.relation
        if not index is None:
            parent_data[index] = mb_data
        elif len(attribute_key) == 0:
            self.__json_data = mb_data
        else:
            parent_data[attribute] = mb_data

    def visit_collection(self, attribute_key, attribute, collection_node,
                         collection_data, is_link_node, parent_data):
        if is_link_node:
            coll_data = collection_node.get_url()
        else:
            coll_data = \
                [mb_data[1] for mb_data in sorted(collection_data.items())]
        if len(attribute_key) == 0:
            self.__json_data = coll_data
        else:
            parent_data[attribute] = coll_data

    @property
    def json_data(self):
        return self.__json_data


class JsonRepresentationGenerator(RepresentationGenerator):
    """
    A JSON generator for resource data.
    """
    def run(self, data_element):
        trv = DataElementTreeTraverser(data_element, self._mapping)
        vst = JsonDataElementTreeVisitor()
        trv.run(vst)
        rpr_string = dumps(vst.json_data)
        self._stream.write(rpr_string)


class JsonResourceRepresenter(MappingResourceRepresenter):
    """
    Resource representer implementation for JSON.
    """
    content_type = JsonMime

    @classmethod
    def make_mapping_registry(cls):
        return JsonMappingRegistry()

    def _make_representation_parser(self, stream, resource_class, mapping):
        parser = JsonRepresentationParser(stream, resource_class, mapping)
        parser.set_option('encoding', self.encoding)
        return parser

    def _make_representation_generator(self, stream, resource_class, mapping):
        generator = JsonRepresentationGenerator(stream, resource_class, mapping)
        generator.set_option('encoding', self.encoding)
        return generator


class JsonMemberDataElement(SimpleMemberDataElement):
    converter_registry = JsonConverterRegistry


class JsonCollectionDataElement(SimpleCollectionDataElement):
    pass


class JsonLinkedDataElement(SimpleLinkedDataElement):
    pass


class JsonRepresenterConfiguration(RepresenterConfiguration):
    pass


class JsonMappingRegistry(SimpleMappingRegistry):
    """
    Registry for JSON mappings.
    """
    member_data_element_base_class = JsonMemberDataElement
    collection_data_element_base_class = JsonCollectionDataElement
    linked_data_element_base_class = JsonLinkedDataElement
    configuration_class = JsonRepresenterConfiguration



"""
Message notification and handling.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 24, 2011.
"""
from everest.entities.system import UserMessage
from everest.interfaces import IUserMessage
from everest.interfaces import IUserMessageChecker
from everest.interfaces import IUserMessageNotifier
from pyramid.threadlocal import get_current_registry
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['MESSENGER_KINDS',
           'UserMessageChecker',
           'UserMessageNotifier',
           'UserMessageHandlingContextManager',
           ]


class MESSENGER_KINDS(object):
    TRANSIENT = 'transient'
    PERSISTENT = 'persistent'


@implementer(IUserMessageNotifier)
class UserMessageNotifier(object):
    """
    Notifier for user messages.
    """
    def notify(self, message_text):
        msg = UserMessage(message_text)
        reg = get_current_registry()
        vote = True
        checkers = reg.subscribers([msg], IUserMessageChecker)
        for checker in checkers:
            vote = checker.check()
            if vote is False:
                # No further processing.
                break
            elif vote is None:
                # Unconditional further processing.
                vote = True
                break
        # Inform all checkers of the final vote.
        for checker in checkers:
            checker.vote = vote
        return vote


@implementer(IUserMessageChecker)
class UserMessageChecker(object):
    """
    Abstract base class for user message checkers.

    User message checkers can be used to decide if further processing
    should be stopped in response to a non-critical event reported through
    a user message.
    """
    def __init__(self):
        self.__message = None
        # The default vote is True, i.e., continue processing.
        self.vote = True

    def __call__(self, message):
        self.vote = True
        self.__message = message
        return self

    @property
    def message(self):
        return self.__message

    def check(self):
        raise NotImplementedError('Abstract method.')


class UserMessageHandlingContextManager(object):
    """
    A context which sets up a user message checker as a subscriber to
    user messages.
    """
    def __init__(self, checker):
        """
        Constructor.

        :param checker: The user message checker to subscribe to user messages.
        :type checker: :class:`everest.messaging.UserMessageChecker` instance.
        """
        self.__checker = checker

    def __enter__(self):
        reg = get_current_registry()
        reg.registerSubscriptionAdapter(self.__checker, (IUserMessage,),
                                        IUserMessageChecker)

    def __exit__(self, ext_type, value, tb):
        reg = get_current_registry()
        reg.unregisterSubscriptionAdapter(self.__checker, (IUserMessage,),
                                          IUserMessageChecker)



"""
Order specification visitor classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 5, 2011.
"""
from everest.entities.utils import slug_from_identifier
from everest.querying.base import CqlExpression
from everest.querying.base import SpecificationVisitor
from everest.querying.interfaces import IOrderSpecificationVisitor
from everest.querying.operators import CQL_ORDER_OPERATORS
from functools import reduce as func_reduce
from operator import and_ as and_operator
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['BubbleSorter',
           'CqlOrderExpression',
           'CqlOrderSpecificationVisitor',
           'OrderSpecificationVisitor',
           'RepositoryOrderSpecificationVisitor',
           'Sorter',
           'SorterTemplate',
           ]


class Sorter(object):
    _order = None

    def __init__(self, order):
        if self.__class__ is Sorter:
            raise NotImplementedError('Abstract class')
        self._order = order

    def sort(self, lst, lo=None, hi=None):
        raise NotImplementedError('Abstract method')

    def set_order(self, order):
        self._order = order


class SorterTemplate(Sorter):
    def __init__(self, order):
        if self.__class__ is SorterTemplate:
            raise NotImplementedError('Abstract class')
        Sorter.__init__(self, order)

    def sort(self, lst, lo=None, hi=None):
        lo = 0 if lo is None else lo
        hi = (len(lst) - 1) if hi is None else hi
        if lo < hi:
            s = self._split(lst, lo, hi)
            self.sort(lst, lo, s - 1)
            self.sort(lst, s, hi)
            self._join(lst, lo, s, hi)

    def _split(self, lst, lo, hi):
        raise NotImplementedError('Abstract method')

    def _join(self, lst, lo, s, hi):
        raise NotImplementedError('Abstract method')


class BubbleSorter(SorterTemplate):
    def __init__(self, order):
        SorterTemplate.__init__(self, order)

    def _split(self, lst, lo, hi):
        j = hi
        while lo < j:
            if self._order.lt(lst[j], lst[j - 1]):
                temp = lst[j]
                lst[j] = lst[j - 1]
                lst[j - 1] = temp
            j -= 1
        return lo + 1

    def _join(self, lst, low_index, split_index, high_index):
        pass


class OrderSpecificationVisitor(SpecificationVisitor):
    """
    Base class for order specification visitors.
    """
    def _asc_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _desc_op(self, spec):
        raise NotImplementedError('Abstract method.')


class CqlOrderExpression(CqlExpression):
    """
    CQL expression representing an order criterion.
    """
    __cql_format = '%(attr)s:%(op)s'

    def __init__(self, attr_name, op_name):
        CqlExpression.__init__(self)
        self.attr_name = attr_name
        self.op_name = op_name

    def _as_string(self):
        return self.__cql_format % dict(attr=self.attr_name,
                                        op=self.op_name)


@implementer(IOrderSpecificationVisitor)
class CqlOrderSpecificationVisitor(OrderSpecificationVisitor):
    """
    Order specification visitor building a CQL expression.
    """
    def _conjunction_op(self, spec, *expressions):
        res = func_reduce(and_operator, expressions)
        return res

    def _asc_op(self, spec):
        return CqlOrderExpression(self.__preprocess_attribute(spec.attr_name),
                                  CQL_ORDER_OPERATORS.ASCENDING.name)

    def _desc_op(self, spec):
        return CqlOrderExpression(self.__preprocess_attribute(spec.attr_name),
                                  CQL_ORDER_OPERATORS.DESCENDING.name)

    def __preprocess_attribute(self, attr_name):
        return slug_from_identifier(attr_name)


class RepositoryOrderSpecificationVisitor(OrderSpecificationVisitor): # pylint: disable=W0223
    """
    Specification visitors that build order expressions for a repository
    backend.
    """
    def __init__(self, entity_class):
        OrderSpecificationVisitor.__init__(self)
        self._entity_class = entity_class

    def order_query(self, query):
        """
        Returns the given query ordered by this visitor's order expression.
        """
        return query.order(self.expression)



"""
View related utilities.

This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 4, 2011.
"""
from everest.mime import CSV_MIME

__docformat__ = 'reStructuredText en'
__all__ = ['accept_csv_only',
           ]


def accept_csv_only(context, request): # pylint: disable-msg=W0613
    """
    This can be used as a custom predicate for view configurations with a
    CSV renderer that should only be invoked if this has been explicitly
    requested in the ACCEPT header by the client.
    """
    return CSV_MIME in [acc.lower() for acc in request.accept]



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 4, 2011.
"""
from everest.querying.ordering import BubbleSorter
from everest.querying.specifications import AscendingOrderSpecification
from everest.querying.specifications import DescendingOrderSpecification
from everest.repositories.rdb import SqlOrderSpecificationVisitor
from everest.testing import Pep8CompliantTestCase
from everest.tests.simple_app.entities import FooEntity
from mock import MagicMock
import random

__docformat__ = 'reStructuredText en'
__all__ = ['AscendingOrderSpecificationTestCase',
           'ConjunctionOrderSpecificationTestCase',
           'ReverseOrderSpecificationTestCase',
           'SorterTestCase',
           'SqlOrderingSpecificationVisitorTestCase',
           ]


class Person(object):

    name = None
    age = None

    def __init__(self, name, age):
        self.name = name
        self.age = age

    def __repr__(self):
        str_format = '<%s name: %s, age: %s>'
        params = (self.__class__.__name__, self.name, self.age)
        return str_format % params


class OrderSpecificationTestCase(Pep8CompliantTestCase):

    def set_up(self):
        self.person_a1 = Person('John', 33)
        self.person_a2 = Person('John', 44)
        self.person_b1 = Person('Mike', 55)
        self.person_b2 = Person('Mike', 66)

    def create_asc_order(self, attr_name):
        return AscendingOrderSpecification(attr_name)

    def create_desc_order(self, attr_name):
        return DescendingOrderSpecification(attr_name)


class AscendingOrderSpecificationTestCase(OrderSpecificationTestCase):

    def test_order_is_satisfied(self):
        order = self.create_asc_order('age')
        self.assert_true(order.lt(self.person_a1, self.person_b1))

    def test_order_is_not_satisfied(self):
        order = self.create_asc_order('age')
        self.assert_false(order.lt(self.person_b1, self.person_a1))


class ReverseOrderSpecificationTestCase(OrderSpecificationTestCase):

    def test_order_is_satisfied(self):
        order = self.create_desc_order('age')
        self.assert_true(order.lt(self.person_b1, self.person_a1))

    def test_order_is_not_satisfied(self):
        order = self.create_desc_order('age')
        self.assert_false(order.lt(self.person_a1, self.person_b1))


class ConjunctionOrderSpecificationTestCase(OrderSpecificationTestCase):

    def test_order_is_satisfied(self):
        order = self.create_asc_order('name') & self.create_asc_order('age')
        self.assert_true(order.lt(self.person_a1, self.person_b1))

    def test_order_is_not_satisfied(self):
        order = self.create_asc_order('name') & self.create_desc_order('age')
        self.assert_false(order.lt(self.person_b1, self.person_a1))

    def test_reverse_order_is_satisfied(self):
        order = self.create_asc_order('name') & self.create_desc_order('age')
        self.assert_true(order.lt(self.person_a2, self.person_a1))

    def test_reverse_order_is_not_satisfied(self):
        order = self.create_asc_order('name') & self.create_desc_order('age')
        self.assert_false(order.lt(self.person_a1, self.person_a2))


class SorterTestCase(Pep8CompliantTestCase):
    andrew11 = None
    bill22 = None
    john33 = None
    john44 = None
    mike55 = None
    mike66 = None
    persons = None

    def set_up(self):
        self.andrew11 = Person('Andrew', 11)
        self.bill22 = Person('Bill', 22)
        self.john33 = Person('John', 33)
        self.john44 = Person('John', 44)
        self.mike55 = Person('Mike', 55)
        self.mike66 = Person('Mike', 66)
        self.persons = [self.andrew11, self.bill22, self.john33, self.john44,
                        self.mike55, self.mike66]
        random.shuffle(self.persons)

    def tear_down(self):
        pass

    def create_asc_order(self, attr_name):
        return AscendingOrderSpecification(attr_name)

    def create_desc_order(self, attr_name):
        return DescendingOrderSpecification(attr_name)

    def get_persons_ordered_by_name_and_age(self):
        return [self.andrew11, self.bill22, self.john33, self.john44,
                self.mike55, self.mike66]

    def get_persons_ordered_by_name_reversed_and_age(self):
        return [self.mike55, self.mike66, self.john33, self.john44,
                self.bill22, self.andrew11]

    def get_persons_ordered_by_name_and_age_reversed(self):
        return [self.andrew11, self.bill22, self.john44, self.john33,
                self.mike66, self.mike55]

    def get_persons_reversed_ordered_by_name_and_age(self):
        return [self.mike66, self.mike55, self.john44, self.john33,
                self.bill22, self.andrew11]

    def test_bubble_sorter(self):
        order = self.create_asc_order('name') & self.create_asc_order('age')
        sorter = BubbleSorter(order)
        sorter.sort(self.persons)
        self.assert_equal(self.persons,
                          self.get_persons_ordered_by_name_and_age())

        order = self.create_desc_order('name') & self.create_asc_order('age')
        sorter.set_order(order)
        sorter.sort(self.persons)
        self.assert_equal(self.persons,
                          self.get_persons_ordered_by_name_reversed_and_age())

        order = self.create_asc_order('name') & self.create_desc_order('age')
        sorter.set_order(order)
        sorter.sort(self.persons)
        self.assert_equal(self.persons,
                          self.get_persons_ordered_by_name_and_age_reversed())

        order = self.create_desc_order('name') & self.create_desc_order('age')
        sorter.set_order(order)
        sorter.sort(self.persons)
        self.assert_equal(self.persons,
                          self.get_persons_reversed_ordered_by_name_and_age())


class SqlOrderingSpecificationVisitorTestCase(Pep8CompliantTestCase):
    class MySqlOrderSpecificationVisitor(SqlOrderSpecificationVisitor):
        def _asc_op(self, spec):
            return lambda value: None

    def test_custom_clause(self):
        obj = object()
        spec = AscendingOrderSpecification('id')
        join_map = {'id':[obj]}
        def get_item(key):
            return join_map[key]
        def contains(key):
            return key in join_map
        mock_join_map = MagicMock(spec=dict)
        mock_join_map.__getitem__.side_effect = get_item
        mock_join_map.__contains__.side_effect = contains
        visitor = self.MySqlOrderSpecificationVisitor(FooEntity,
                                                      custom_join_clauses=
                                                                mock_join_map)
        visitor.visit_nullary(spec)
        mock_join_map.__getitem__.assert_called_with('id')



"""
System entities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 22, 2011.
"""
from datetime import datetime
from everest.entities.base import Entity
from everest.interfaces import IUserMessage
from zope.interface import implementer # pylint: disable=E0611,F0401
import uuid

__docformat__ = 'reStructuredText en'
__all__ = ['UserMessage',
           ]


@implementer(IUserMessage)
class UserMessage(Entity):
    """
    A user message holding a text, a GUID and a time stamp.
    """
    def __init__(self, text, guid=None, time_stamp=None, **kw):
        msg_id = kw.pop('id', None)
        if msg_id is None:
            if guid is None:
                guid = str(uuid.uuid4())
            msg_id = guid
        elif not guid is None and msg_id != guid:
            raise ValueError('Can not pass different values for "guid" and '
                             '"id" parameter.')
        else:
            guid = msg_id
        kw['id'] = msg_id
        Entity.__init__(self, **kw)
        if time_stamp is None:
            time_stamp = datetime.now()
        self.text = text
        self.guid = guid
        self.time_stamp = time_stamp

    @property
    def slug(self):
        return self.guid
